"""Pulsar bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "PulsarChannelBindings",
    "PulsarMessageBindings",
    "PulsarOperationBindings",
    "PulsarRetention",
    "PulsarServerBindings",
]

from typing import Literal

from pydantic import Field

from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null


class PulsarServerBindings(NonExtendableBaseModel):
    """
    Pulsar Server Binding Object.

    This object contains information about the server representation in Pulsar.
    """

    tenant: str = Field(
        default="public",
        description="The pulsar tenant. If omitted, 'public' MUST be assumed.",
    )
    binding_version: str = Field(
        default="0.1.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )


class PulsarRetention(NonExtendableBaseModel):
    """
    Retention Definition Object.

    The Retention Definition Object is used to describe the Pulsar Retention policy.
    """

    time: int | None = Field(
        default=0,
        exclude_if=is_null,
        description="Time given in Minutes.",
    )
    size: int | None = Field(
        default=0,
        exclude_if=is_null,
        description="Size given in MegaBytes.",
    )


class PulsarChannelBindings(NonExtendableBaseModel):
    """
    Pulsar Channel Binding Object.

    This object contains information about the channel representation in Pulsar.
    """

    namespace: str = Field(
        ...,
        description="The namespace the channel is associated with.",
    )
    persistence: Literal["persistent", "non-persistent"] = Field(
        ...,
        description=(
            "Persistence of the topic in Pulsar. It MUST be either persistent or "
            "non-persistent."
        ),
    )
    compaction: int | None = Field(
        default=None,
        exclude_if=is_null,
        description="Topic compaction threshold given in Megabytes.",
    )
    geo_replication: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="geo-replication",
        description="A list of clusters the topic is replicated to.",
    )
    retention: PulsarRetention | None = Field(
        default=None,
        exclude_if=is_null,
        description="Topic retention policy.",
    )
    ttl: int | None = Field(
        default=None,
        exclude_if=is_null,
        description="Message time-to-live in seconds.",
    )
    deduplication: bool | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Message deduplication. When true, it ensures that each message produced "
            "on Pulsar topics is persisted to disk only once."
        ),
    )
    binding_version: str = Field(
        default="0.1.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )


class PulsarOperationBindings(NonExtendableBaseModel):
    """
    Pulsar Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class PulsarMessageBindings(NonExtendableBaseModel):
    """
    Pulsar Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""AMQP 1.0 bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "AMQP1ChannelBindings",
    "AMQP1MessageBindings",
    "AMQP1OperationBindings",
    "AMQP1ServerBindings",
]

from asyncapi3.models.base_models import NonExtendableBaseModel


class AMQP1ServerBindings(NonExtendableBaseModel):
    """
    AMQP 1.0 Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class AMQP1ChannelBindings(NonExtendableBaseModel):
    """
    AMQP 1.0 Channel Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class AMQP1OperationBindings(NonExtendableBaseModel):
    """
    AMQP 1.0 Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class AMQP1MessageBindings(NonExtendableBaseModel):
    """
    AMQP 1.0 Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""AsyncAPI root model for AsyncAPI 3.0 specification."""

__all__ = ["AsyncAPI3"]

import re

from pydantic import AnyUrl, Field, field_validator

from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.channel import Channels
from asyncapi3.models.components import Components
from asyncapi3.models.helpers import is_null
from asyncapi3.models.info import Info
from asyncapi3.models.operation import Operations
from asyncapi3.models.server import Servers


class AsyncAPI3(ExtendableBaseModel):
    """
    AsyncAPI Object.

    This is the root document object for the AsyncAPI 3.0 specification.
    It combines resource listing and API declaration into one document.

    This object MAY be extended with Specification Extensions.
    """

    asyncapi: str = Field(
        default="3.0.0",
        description=(
            "Specifies the AsyncAPI Specification version being used. It can be used "
            "by tooling Specifications and clients to interpret the version. "
            "The structure shall be major.minor.patch, where patch versions must be "
            "compatible with the existing major.minor tooling. Typically patch "
            "versions will be introduced to address errors in the documentation, "
            "and tooling should typically be compatible with the corresponding "
            "major.minor (1.0.*). Patch versions will correspond to patches of this "
            "document."
        ),
    )
    id: AnyUrl | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Identifier of the application the AsyncAPI document is defining. "
            "This field represents a unique universal identifier of the application "
            "the AsyncAPI document is defining. It must conform to the URI format, "
            "according to RFC3986. It is RECOMMENDED to use a URN to globally and "
            "uniquely identify the application during long periods of time, even "
            "after it becomes unavailable or ceases to exist."
        ),
    )
    info: Info = Field(
        description=(
            "Provides metadata about the API. The metadata can be used by the clients "
            "if needed."
        ),
    )
    servers: Servers | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The Servers Object is a map of Server Objects. Provides connection "
            "details of servers."
        ),
    )
    default_content_type: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="defaultContentType",
        description=(
            "Default content type to use when encoding/decoding a message's payload."
        ),
    )
    channels: Channels | None = Field(
        default=None,
        exclude_if=is_null,
        description="The channels used by this application.",
    )
    operations: Operations | None = Field(
        default=None,
        exclude_if=is_null,
        description="The operations this application MUST implement.",
    )
    components: Components | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An element to hold various reusable objects for the specification. "
            "Everything that is defined inside this object represents a resource "
            "that MAY or MAY NOT be used in the rest of the document and MAY or "
            "MAY NOT be used by the implemented Application."
        ),
    )

    @field_validator("asyncapi")
    @classmethod
    def validate_asyncapi_version(cls, version: str) -> str:
        """Validate AsyncAPI version format (semantic versioning) and compatibility."""
        # Pattern for semantic versioning: major.minor.patch[-suffix]
        pattern = r"^\d+\.\d+\.\d+(-[\w\.\-]+)?$"
        if not re.match(pattern, version):
            raise ValueError(
                "asyncapi must be in semantic versioning format: "
                "major.minor.patch[-suffix]"
            )

        # Check that major version is 3 (this application only supports AsyncAPI 3.x.x)
        major_version = int(version.split(".")[0])
        if major_version != 3:
            raise ValueError(
                f"asyncapi version {version} is not supported. "
                "This application only supports AsyncAPI 3.x.x versions"
            )

        return version



"""Redis bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "RedisChannelBindings",
    "RedisMessageBindings",
    "RedisOperationBindings",
    "RedisServerBindings",
]

from asyncapi3.models.base_models import NonExtendableBaseModel


class RedisServerBindings(NonExtendableBaseModel):
    """
    Redis Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class RedisChannelBindings(NonExtendableBaseModel):
    """
    Redis Channel Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class RedisOperationBindings(NonExtendableBaseModel):
    """
    Redis Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class RedisMessageBindings(NonExtendableBaseModel):
    """
    Redis Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""Schema models for AsyncAPI 3.0 specification."""

__all__ = [
    "MultiFormatSchema",
    "Schema",
]
from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from asyncapi3.models.base import ExternalDocumentation, Reference
from asyncapi3.models.helpers import is_null


# TODO: Not full at this point. Should go to dict instead?
# TODO: MAY be represented by the boolean value true
class Schema(BaseModel):
    """
    Schema Object.

    The Schema Object allows the definition of input and output data types.
    These types can be objects, but also primitives and arrays. This object is a
    superset of the JSON Schema Specification Draft 07. The empty schema (which allows
    any instance to validate) MAY be represented by the boolean value true and a schema
    which allows no instance to validate MAY be represented by the boolean value false.

    Further information about the properties can be found in JSON Schema Core and JSON
    Schema Validation. Unless stated otherwise, the property definitions follow the
    JSON Schema specification as referenced here. For other formats (e.g., Avro, RAML,
    etc) see Multi Format Schema Object.

    The AsyncAPI Schema Object is a JSON Schema vocabulary which extends JSON Schema
    Core and Validation vocabularies. As such, any keyword available for those
    vocabularies is by definition available in AsyncAPI, and will work the exact same
    way, including but not limited to: title, type, required, multipleOf, maximum,
    exclusiveMaximum, minimum, exclusiveMinimum, maxLength, minLength, pattern,
    maxItems, minItems, uniqueItems, maxProperties, minProperties, enum, const,
    examples, if/then/else, readOnly, writeOnly, properties, patternProperties,
    additionalProperties, additionalItems, items, propertyNames, contains, allOf,
    oneOf, anyOf, not.

    The following properties are taken from the JSON Schema definition but their
    definitions were adjusted to the AsyncAPI Specification:

    - description - CommonMark syntax can be used for rich text representation.
    - format - See Data Type Formats for further details. While relying on JSON
      Schema's defined formats, the AsyncAPI Specification offers a few additional
      predefined formats.
    - default - Use it to specify that property has a predefined value if no other
      value is present. Unlike JSON Schema, the value MUST conform to the defined type
      for the Schema Object defined at the same level. For example, of type is string,
      then default can be "foo" but cannot be 1.

    Alternatively, any time a Schema Object can be used, a Reference Object can be
    used in its place. This allows referencing definitions in place of defining them
    inline. It is appropriate to clarify that the $ref keyword MUST follow the behavior
    described by Reference Object instead of the one in JSON Schema definition.

    In addition to the JSON Schema fields, the following AsyncAPI vocabulary fields
    MAY be used for further schema documentation:

    This object MAY be extended with Specification Extensions.
    """

    # JSON Schema fields - these are dynamic and can be any JSON Schema properties
    model_config = ConfigDict(
        extra="allow",
        revalidate_instances="always",
        validate_assignment=True,
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )

    discriminator: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Adds support for polymorphism. The discriminator is the schema property "
            "name that is used to differentiate between other schema that inherit this "
            "schema. The property name used MUST be defined at this schema and it MUST "
            "be in the required property list. When used, the value MUST be the name "
            "of this schema or any schema that inherits it. See Composition and "
            "Inheritance for more details."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description=("Additional external documentation for this schema."),
    )
    deprecated: bool | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Specifies that a schema is deprecated and SHOULD be transitioned out of "
            "usage. Default value is false."
        ),
    )


class MultiFormatSchema(BaseModel):
    """
    Multi Format Schema Object.

    The Multi Format Schema Object represents a schema definition. It differs from the
    Schema Object in that it supports multiple schema formats or languages (e.g., JSON
    Schema, Avro, etc.).

    This object MAY be extended with Specification Extensions.
    """

    model_config = ConfigDict(
        extra="allow",
        revalidate_instances="always",
        validate_assignment=True,
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )

    schema_format: str = Field(
        alias="schemaFormat",
        default="application/vnd.aai.asyncapi;version=3.0.0",
        description=(
            "REQUIRED. A string containing the name of the schema format that is "
            "used to define the information. If schemaFormat is missing, it MUST "
            "default to application/vnd.aai.asyncapi+json;version={{asyncapi}} where "
            "{{asyncapi}} matches the AsyncAPI Version String. In such a case, this "
            "would make the Multi Format Schema Object equivalent to the Schema "
            "Object. When using Reference Object within the schema, the schemaFormat "
            "of the resource being referenced MUST match the schemaFormat of the "
            "schema that contains the initial reference. For example, if you "
            "reference Avro schema, then schemaFormat of referencing resource and "
            "the resource being reference MUST match. Check out the supported "
            "schema formats table for more information. Custom values are allowed "
            "but their implementation is OPTIONAL. A custom value MUST NOT refer "
            "to one of the schema formats listed in the table. When using Reference "
            "Objects within the schema, the schemaFormat of the referenced resource "
            "MUST match the schemaFormat of the schema containing the reference."
        ),
    )
    # TODO: So either Schema, dict or string?
    schema: Any = Field(
        description=(
            "REQUIRED. Definition of the message payload. It can be of any type but "
            "defaults to Schema Object. It MUST match the schema format defined in "
            "schemaFormat, including the encoding type. E.g., Avro should be inlined "
            "as either a YAML or JSON object instead of as a string to be parsed as "
            "YAML or JSON. Non-JSON-based schemas (e.g., Protobuf or XSD) MUST be "
            "inlined as a string."
        ),
    )



"""JMS bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "JMSChannelBindings",
    "JMSMessageBindings",
    "JMSOperationBindings",
    "JMSServerBindings",
]

from typing import Literal

from pydantic import Field

from asyncapi3.models.base import Reference
from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null
from asyncapi3.models.schema import Schema


class JMSServerBindings(NonExtendableBaseModel):
    """
    JMS Server Binding Object.

    The JMS Server Binding Object is defined by a JSON Schema, which defines these
    fields.
    """

    jms_connection_factory: str = Field(
        alias="jmsConnectionFactory",
        description=(
            "REQUIRED. The classname of the ConnectionFactory implementation for "
            "the JMS Provider."
        ),
    )
    properties: list[Schema] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Additional properties to set on the JMS ConnectionFactory implementation "
            "for the JMS Provider."
        ),
    )
    client_id: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="clientID",
        description=(
            "A client identifier for applications that use this JMS connection factory "
            "If the Client ID Policy is set to 'Restricted' (the default), then "
            "configuring a Client ID on the ConnectionFactory prevents more than one "
            "JMS client from using a connection from this factory."
        ),
    )
    binding_version: str = Field(
        default="0.0.1",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )


class JMSChannelBindings(NonExtendableBaseModel):
    """
    JMS Channel Binding Object.

    The JMS Channel Binding Object is defined by a JSON Schema, which defines these
    fields.
    """

    destination: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The destination (queue) name for this channel. SHOULD only be specified "
            "if the channel name differs from the actual destination name, such as "
            "when the channel name is not a valid destination name according to the "
            "JMS Provider. Optional, defaults to the channel name."
        ),
    )
    destination_type: Literal["queue", "fifo-queue"] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="destinationType",
        description=(
            "The type of destination, which MUST be either queue, or fifo-queue. "
            "SHOULD be specified to document the messaging model (point-to-point, "
            "or strict message ordering) supported by this channel. Optional, defaults "
            "to queue."
        ),
    )
    binding_version: str = Field(
        default="0.0.1",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )


class JMSOperationBindings(NonExtendableBaseModel):
    """
    JMS Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class JMSMessageBindings(NonExtendableBaseModel):
    """
    JMS Message Binding Object.

    The JMS Message Binding Object is defined by a JSON Schema, which defines these
    fields.
    """

    headers: Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A Schema object containing the definitions for JMS specific headers "
            "(so-called protocol headers). This schema MUST be of type object and "
            "have a properties key. Examples of JMS protocol headers are JMSMessageID, "
            "JMSTimestamp, and JMSCorrelationID."
        ),
    )
    binding_version: str = Field(
        default="0.0.1",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )



"""Components model for AsyncAPI 3.0 specification."""

__all__ = [
    "ChannelBindings",
    "Channels",
    "Components",
    "CorrelationIDs",
    "ExternalDocs",
    "MessageBindings",
    "Messages",
    "OperationBindings",
    "OperationTraits",
    "Operations",
    "Parameters",
    "Replies",
    "ReplyAddresses",
    "Schemas",
    "SecuritySchemes",
    "ServerBindings",
    "ServerVariables",
    "Servers",
    "Tags",
]


from pydantic import Field

from asyncapi3.models.base import ExternalDocumentation, Reference, Tag
from asyncapi3.models.base_models import ExtendableBaseModel, PatternedRootModel
from asyncapi3.models.bindings import (
    ChannelBindingsObject,
    MessageBindingsObject,
    OperationBindingsObject,
    ServerBindingsObject,
)
from asyncapi3.models.channel import Channel, Parameter
from asyncapi3.models.helpers import is_null
from asyncapi3.models.message import Message, MessageTrait
from asyncapi3.models.operation import (
    Operation,
    OperationReply,
    OperationReplyAddress,
    OperationTrait,
)
from asyncapi3.models.schema import MultiFormatSchema, Schema
from asyncapi3.models.security import CorrelationID, SecurityScheme
from asyncapi3.models.server import Server, ServerVariable


class Schemas(PatternedRootModel[MultiFormatSchema | Schema | Reference]):
    """
    Schemas Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or Schema objects.
    """


class Servers(PatternedRootModel[Server | Reference]):
    """
    Servers Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or Server objects.
    """


class Channels(PatternedRootModel[Channel | Reference]):
    """
    Channels Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or Channel objects.
    """


class Operations(PatternedRootModel[Operation | Reference]):
    """
    Operations Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or Operation objects.
    """


class Messages(PatternedRootModel[Message | Reference]):
    """
    Messages Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or Message objects.
    """


class SecuritySchemes(PatternedRootModel[SecurityScheme | Reference]):
    """
    SecuritySchemes Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or SecurityScheme objects.
    """


class ServerVariables(PatternedRootModel[ServerVariable | Reference]):
    """
    ServerVariable Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or ServerVariable objects.
    """


class Parameters(PatternedRootModel[Parameter | Reference]):
    """
    Parameter Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or Parameter objects.
    """


class CorrelationIDs(PatternedRootModel[CorrelationID | Reference]):
    """
    CorrelationID Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or CorrelationID objects.
    """


class Replies(PatternedRootModel[OperationReply | Reference]):
    """
    OperationReply Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or OperationReply objects.
    """


class ReplyAddresses(PatternedRootModel[OperationReplyAddress | Reference]):
    """
    OperationReplyAddress Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or OperationReplyAddress objects.
    """


class ExternalDocs(PatternedRootModel[ExternalDocumentation | Reference]):
    """
    ExternalDocumentation Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or ExternalDocumentation objects.
    """


class Tags(PatternedRootModel[Tag | Reference]):
    """
    Tag Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or Tag objects.
    """


class OperationTraits(PatternedRootModel[OperationTrait | Reference]):
    """
    OperationTrait Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or OperationTrait objects.
    """


class MessageTraits(PatternedRootModel[MessageTrait | Reference]):
    """
    MessageTrait Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or MessageTrait objects.
    """


class ServerBindings(PatternedRootModel[ServerBindingsObject | Reference]):
    """
    ServerBindingsObject Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or ServerBindingsObject objects.
    """


class ChannelBindings(PatternedRootModel[ChannelBindingsObject | Reference]):
    """
    ChannelBindingsObject Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or ChannelBindingsObject objects.
    """


class OperationBindings(PatternedRootModel[OperationBindingsObject | Reference]):
    """
    OperationBindingsObject Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or OperationBindingsObject objects.
    """


class MessageBindings(PatternedRootModel[MessageBindingsObject | Reference]):
    """
    MessageBindingsObject Object.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$, values match Reference or MessageBindingsObject objects.
    """


class Components(ExtendableBaseModel):
    """
    Components Object.

    Holds a set of reusable objects for different aspects of the AsyncAPI specification.
    All objects defined within the components object will have no effect on the API
    unless they are explicitly referenced from properties outside the components object.

    This object MAY be extended with Specification Extensions.
    """

    schemas: Schemas | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An object to hold reusable Schema Object. If this is a Schema Object, "
            "then the schemaFormat will be assumed to be "
            "'application/vnd.aai.asyncapi+json;version=asyncapi' where the version "
            "is equal to the AsyncAPI Version String."
        ),
    )
    servers: Servers | None = Field(
        default=None,
        exclude_if=is_null,
        description="An object to hold reusable Server Objects.",
    )
    channels: Channels | None = Field(
        default=None,
        exclude_if=is_null,
        description="An object to hold reusable Channel Objects.",
    )
    operations: Operations | None = Field(
        default=None,
        exclude_if=is_null,
        description="An object to hold reusable Operation Objects.",
    )
    messages: Messages | None = Field(
        default=None,
        exclude_if=is_null,
        description="An object to hold reusable Message Objects.",
    )
    security_schemes: SecuritySchemes | None = Field(
        default=None,
        exclude_if=is_null,
        alias="securitySchemes",
        description="An object to hold reusable Security Scheme Objects.",
    )
    server_variables: ServerVariables | None = Field(
        default=None,
        exclude_if=is_null,
        alias="serverVariables",
        description="An object to hold reusable Server Variable Objects.",
    )
    parameters: Parameters | None = Field(
        default=None,
        exclude_if=is_null,
        description="An object to hold reusable Parameter Objects.",
    )
    correlation_ids: CorrelationIDs | None = Field(
        default=None,
        exclude_if=is_null,
        alias="correlationIds",
        description="An object to hold reusable Correlation ID Objects.",
    )
    replies: Replies | None = Field(
        default=None,
        exclude_if=is_null,
        description="An object to hold reusable Operation Reply Objects.",
    )
    reply_addresses: ReplyAddresses | None = Field(
        default=None,
        exclude_if=is_null,
        alias="replyAddresses",
        description="An object to hold reusable Operation Reply Address Objects.",
    )
    external_docs: ExternalDocs | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="An object to hold reusable External Documentation Objects.",
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description="An object to hold reusable Tag Objects.",
    )
    operation_traits: OperationTraits | None = Field(
        default=None,
        exclude_if=is_null,
        alias="operationTraits",
        description="An object to hold reusable Operation Trait Objects.",
    )
    message_traits: MessageTraits | None = Field(
        default=None,
        exclude_if=is_null,
        alias="messageTraits",
        description="An object to hold reusable Message Trait Objects.",
    )
    server_bindings: ServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        alias="serverBindings",
        description="An object to hold reusable Server Bindings Objects.",
    )
    channel_bindings: ChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        alias="channelBindings",
        description="An object to hold reusable Channel Bindings Objects.",
    )
    operation_bindings: OperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        alias="operationBindings",
        description="An object to hold reusable Operation Bindings Objects.",
    )
    message_bindings: MessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        alias="messageBindings",
        description="An object to hold reusable Message Bindings Objects.",
    )



"""Info models for AsyncAPI 3.0 specification."""

__all__ = ["Contact", "Info", "License"]

from pydantic import Field, HttpUrl

from asyncapi3.models.base import ExternalDocumentation, Reference, Tags
from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.helpers import EmailStr, is_null


class Contact(ExtendableBaseModel):
    """
    Contact Object.

    Contact information for the exposed API.
    """

    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="The identifying name of the contact person/organization.",
    )
    url: HttpUrl | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The URL pointing to the contact information. This MUST be in the form "
            "of an absolute URL."
        ),
    )
    email: EmailStr | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The email address of the contact person/organization. MUST be in the "
            "format of an email address."
        ),
    )


class License(ExtendableBaseModel):
    """
    License Object.

    License information for the exposed API.
    """

    name: str = Field(
        description="The license name used for the API.",
    )
    url: HttpUrl | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A URL to the license used for the API. This MUST be in the form of an "
            "absolute URL."
        ),
    )


class Info(ExtendableBaseModel):
    """
    Info Object.

    The object provides metadata about the API.
    The metadata can be used by the clients if needed.
    """

    title: str = Field(
        description="The title of the application.",
    )
    version: str = Field(
        description=(
            "Provides the version of the application API (not to be confused with "
            "the specification version)."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A short description of the application. CommonMark syntax can be used "
            "for rich text representation."
        ),
    )
    terms_of_service: HttpUrl | None = Field(
        default=None,
        exclude_if=is_null,
        alias="termsOfService",
        description=(
            "A URL to the Terms of Service for the API. This MUST be in the form "
            "of an absolute URL."
        ),
    )
    contact: Contact | None = Field(
        default=None,
        exclude_if=is_null,
        description="The contact information for the exposed API.",
    )
    license: License | None = Field(
        default=None,
        exclude_if=is_null,
        description="The license information for the exposed API.",
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of tags for application API documentation control. Tags can be "
            "used for logical grouping of applications."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation of the exposed API.",
    )



"""STOMP bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "STOMPChannelBindings",
    "STOMPMessageBindings",
    "STOMPOperationBindings",
    "STOMPServerBindings",
]

from asyncapi3.models.base_models import NonExtendableBaseModel


class STOMPServerBindings(NonExtendableBaseModel):
    """
    STOMP Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class STOMPChannelBindings(NonExtendableBaseModel):
    """
    STOMP Channel Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class STOMPOperationBindings(NonExtendableBaseModel):
    """
    STOMP Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class STOMPMessageBindings(NonExtendableBaseModel):
    """
    STOMP Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""Mercure bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "MercureChannelBindings",
    "MercureMessageBindings",
    "MercureOperationBindings",
    "MercureServerBindings",
]

from asyncapi3.models.base_models import NonExtendableBaseModel


class MercureServerBindings(NonExtendableBaseModel):
    """
    Mercure Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class MercureChannelBindings(NonExtendableBaseModel):
    """
    Mercure Channel Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class MercureOperationBindings(NonExtendableBaseModel):
    """
    Mercure Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class MercureMessageBindings(NonExtendableBaseModel):
    """
    Mercure Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""HTTP bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "HTTPChannelBindings",
    "HTTPMessageBindings",
    "HTTPOperationBindings",
    "HTTPServerBindings",
]

from typing import Literal

from pydantic import Field

from asyncapi3.models.base import Reference
from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null
from asyncapi3.models.schema import Schema


class HTTPServerBindings(NonExtendableBaseModel):
    """
    HTTP Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class HTTPChannelBindings(NonExtendableBaseModel):
    """
    HTTP Channel Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class HTTPOperationBindings(NonExtendableBaseModel):
    """
    HTTP Operation Binding Object.

    This object contains information about the operation representation in HTTP.

    This object MUST contain only the properties defined below.
    """

    method: Literal[
        "GET",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
        "HEAD",
        "OPTIONS",
        "CONNECT",
        "TRACE",
    ] = Field(
        description=(
            "The HTTP method for the request. Its value MUST be one of GET, POST, PUT, "
            "PATCH, DELETE, HEAD, OPTIONS, CONNECT, and TRACE."
        ),
    )
    query: Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A Schema object containing the definitions for each query parameter. This "
            "schema MUST be of type object and have a properties key."
        ),
    )
    binding_version: str = Field(
        default="0.3.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )


class HTTPMessageBindings(NonExtendableBaseModel):
    """
    HTTP Message Binding Object.

    This object contains information about the message representation in HTTP.

    This object MUST contain only the properties defined below.
    """

    headers: Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A Schema object containing the definitions for HTTP-specific headers. "
            "This schema MUST be of type object and have a properties key."
        ),
    )
    status_code: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="statusCode",
        description=(
            "The HTTP response status code according to RFC 9110. statusCode is only "
            "relevant for messages referenced by the Operation Reply Object, as it "
            "defines the status code for the response. In all other cases, this value "
            "can be safely ignored."
        ),
    )
    binding_version: str = Field(
        default="0.3.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )



"""Core bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "ChannelBindingsObject",
    "MessageBindingsObject",
    "OperationBindingsObject",
    "ServerBindingsObject",
]

from pydantic import Field

from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.bindings.amqp import (
    AMQPChannelBindings,
    AMQPMessageBindings,
    AMQPOperationBindings,
    AMQPServerBindings,
)
from asyncapi3.models.bindings.amqp1 import (
    AMQP1ChannelBindings,
    AMQP1MessageBindings,
    AMQP1OperationBindings,
    AMQP1ServerBindings,
)
from asyncapi3.models.bindings.anypointmq import (
    AnypointMQChannelBindings,
    AnypointMQMessageBindings,
    AnypointMQOperationBindings,
    AnypointMQServerBindings,
)
from asyncapi3.models.bindings.googlepubsub import (
    GooglePubSubChannelBindings,
    GooglePubSubMessageBindings,
    GooglePubSubOperationBindings,
    GooglePubSubServerBindings,
)
from asyncapi3.models.bindings.http import (
    HTTPChannelBindings,
    HTTPMessageBindings,
    HTTPOperationBindings,
    HTTPServerBindings,
)
from asyncapi3.models.bindings.ibmmq import (
    IBMMQChannelBindings,
    IBMMQMessageBindings,
    IBMMQOperationBindings,
    IBMMQServerBindings,
)
from asyncapi3.models.bindings.jms import (
    JMSChannelBindings,
    JMSMessageBindings,
    JMSOperationBindings,
    JMSServerBindings,
)
from asyncapi3.models.bindings.kafka import (
    KafkaChannelBindings,
    KafkaMessageBindings,
    KafkaOperationBindings,
    KafkaServerBindings,
)
from asyncapi3.models.bindings.mercure import (
    MercureChannelBindings,
    MercureMessageBindings,
    MercureOperationBindings,
    MercureServerBindings,
)
from asyncapi3.models.bindings.mqtt import (
    MQTTChannelBindings,
    MQTTMessageBindings,
    MQTTOperationBindings,
    MQTTServerBindings,
)
from asyncapi3.models.bindings.mqtt5 import (
    MQTT5ChannelBindings,
    MQTT5MessageBindings,
    MQTT5OperationBindings,
    MQTT5ServerBindings,
)
from asyncapi3.models.bindings.nats import (
    NATSChannelBindings,
    NATSMessageBindings,
    NATSOperationBindings,
    NATSServerBindings,
)
from asyncapi3.models.bindings.pulsar import (
    PulsarChannelBindings,
    PulsarMessageBindings,
    PulsarOperationBindings,
    PulsarServerBindings,
)
from asyncapi3.models.bindings.redis import (
    RedisChannelBindings,
    RedisMessageBindings,
    RedisOperationBindings,
    RedisServerBindings,
)
from asyncapi3.models.bindings.sns import (
    SNSChannelBindings,
    SNSMessageBindings,
    SNSOperationBindings,
    SNSServerBindings,
)
from asyncapi3.models.bindings.solace import (
    SolaceChannelBindings,
    SolaceMessageBindings,
    SolaceOperationBindings,
    SolaceServerBindings,
)
from asyncapi3.models.bindings.sqs import (
    SQSChannelBindings,
    SQSMessageBindings,
    SQSOperationBindings,
    SQSServerBindings,
)
from asyncapi3.models.bindings.stomp import (
    STOMPChannelBindings,
    STOMPMessageBindings,
    STOMPOperationBindings,
    STOMPServerBindings,
)
from asyncapi3.models.bindings.websockets import (
    WebSocketsChannelBindings,
    WebSocketsMessageBindings,
    WebSocketsOperationBindings,
    WebSocketsServerBindings,
)
from asyncapi3.models.helpers import is_null


class ServerBindingsObject(ExtendableBaseModel):
    """
    Server Bindings Object.

    Map describing protocol-specific definitions for a server.

    This object MAY be extended with Specification Extensions.
    """

    http: HTTPServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an HTTP server.",
    )
    ws: WebSocketsServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a WebSockets server.",
    )
    kafka: KafkaServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Kafka server.",
    )
    anypointmq: AnypointMQServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an Anypoint MQ server.",
    )
    amqp: AMQPServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 0-9-1 server.",
    )
    amqp1: AMQP1ServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 1.0 server.",
    )
    mqtt: MQTTServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT server.",
    )
    mqtt5: MQTT5ServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT 5 server.",
    )
    nats: NATSServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a NATS server.",
    )
    jms: JMSServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a JMS server.",
    )
    sns: SNSServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SNS server.",
    )
    solace: SolaceServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Solace server.",
    )
    sqs: SQSServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SQS server.",
    )
    stomp: STOMPServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a STOMP server.",
    )
    redis: RedisServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Redis server.",
    )
    mercure: MercureServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Mercure server.",
    )
    ibmmq: IBMMQServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an IBM MQ server.",
    )
    googlepubsub: GooglePubSubServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Google Cloud Pub/Sub server.",
    )
    pulsar: PulsarServerBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Pulsar server.",
    )


class ChannelBindingsObject(ExtendableBaseModel):
    """
    Channel Bindings Object.

    Map describing protocol-specific definitions for a channel.

    This object MAY be extended with Specification Extensions.
    """

    http: HTTPChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an HTTP channel.",
    )
    ws: WebSocketsChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a WebSockets channel.",
    )
    kafka: KafkaChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Kafka channel.",
    )
    anypointmq: AnypointMQChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an Anypoint MQ channel.",
    )
    amqp: AMQPChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 0-9-1 channel.",
    )
    amqp1: AMQP1ChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 1.0 channel.",
    )
    mqtt: MQTTChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT channel.",
    )
    mqtt5: MQTT5ChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT 5 channel.",
    )
    nats: NATSChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a NATS channel.",
    )
    jms: JMSChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a JMS channel.",
    )
    sns: SNSChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SNS channel.",
    )
    solace: SolaceChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Solace channel.",
    )
    sqs: SQSChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SQS channel.",
    )
    stomp: STOMPChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a STOMP channel.",
    )
    redis: RedisChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Redis channel.",
    )
    mercure: MercureChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Mercure channel.",
    )
    ibmmq: IBMMQChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an IBM MQ channel.",
    )
    googlepubsub: GooglePubSubChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Google Cloud Pub/Sub channel.",
    )
    pulsar: PulsarChannelBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Pulsar channel.",
    )


class OperationBindingsObject(ExtendableBaseModel):
    """
    Operation Bindings Object.

    Map describing protocol-specific definitions for an operation.

    This object MAY be extended with Specification Extensions.
    """

    http: HTTPOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an HTTP operation.",
    )
    ws: WebSocketsOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a WebSockets operation.",
    )
    kafka: KafkaOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Kafka operation.",
    )
    anypointmq: AnypointMQOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an Anypoint MQ operation.",
    )
    amqp: AMQPOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 0-9-1 operation.",
    )
    amqp1: AMQP1OperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 1.0 operation.",
    )
    mqtt: MQTTOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT operation.",
    )
    mqtt5: MQTT5OperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT 5 operation.",
    )
    nats: NATSOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a NATS operation.",
    )
    jms: JMSOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a JMS operation.",
    )
    sns: SNSOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SNS operation.",
    )
    solace: SolaceOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Solace operation.",
    )
    sqs: SQSOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SQS operation.",
    )
    stomp: STOMPOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a STOMP operation.",
    )
    redis: RedisOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Redis operation.",
    )
    mercure: MercureOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Mercure operation.",
    )
    ibmmq: IBMMQOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an IBM MQ operation.",
    )
    googlepubsub: GooglePubSubOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Protocol-specific information for a Google Cloud Pub/Sub operation."
        ),
    )
    pulsar: PulsarOperationBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Pulsar operation.",
    )


class MessageBindingsObject(ExtendableBaseModel):
    """
    Message Bindings Object.

    Map describing protocol-specific definitions for a message.

    This object MAY be extended with Specification Extensions.
    """

    http: HTTPMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Protocol-specific information for an HTTP message, i.e., a request or a "
            "response."
        ),
    )
    ws: WebSocketsMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a WebSockets message.",
    )
    kafka: KafkaMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Kafka message.",
    )
    anypointmq: AnypointMQMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an Anypoint MQ message.",
    )
    amqp: AMQPMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 0-9-1 message.",
    )
    amqp1: AMQP1MessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an AMQP 1.0 message.",
    )
    mqtt: MQTTMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT message.",
    )
    mqtt5: MQTT5MessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an MQTT 5 message.",
    )
    nats: NATSMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a NATS message.",
    )
    jms: JMSMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a JMS message.",
    )
    sns: SNSMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SNS message.",
    )
    solace: SolaceMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Solace message.",
    )
    sqs: SQSMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an SQS message.",
    )
    stomp: STOMPMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a STOMP message.",
    )
    redis: RedisMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Redis message.",
    )
    mercure: MercureMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Mercure message.",
    )
    ibmmq: IBMMQMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for an IBM MQ message.",
    )
    googlepubsub: GooglePubSubMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Google Cloud Pub/Sub message.",
    )
    pulsar: PulsarMessageBindings | None = Field(
        default=None,
        exclude_if=is_null,
        description="Protocol-specific information for a Pulsar message.",
    )



"""WebSockets bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "WebSocketsChannelBindings",
    "WebSocketsMessageBindings",
    "WebSocketsOperationBindings",
    "WebSocketsServerBindings",
]

from typing import Literal

from pydantic import Field

from asyncapi3.models.base import Reference
from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null
from asyncapi3.models.schema import Schema


class WebSocketsServerBindings(NonExtendableBaseModel):
    """
    WebSockets Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class WebSocketsChannelBindings(NonExtendableBaseModel):
    """
    WebSockets Channel Binding Object.

    When using WebSockets, the channel represents the connection. Unlike other protocols
    that support multiple virtual channels (topics, routing keys, etc.) per connection,
    WebSockets doesn't support virtual channels or, put it another way, there's only
    one channel and its characteristics are strongly related to the protocol used for
    the handshake, i.e., HTTP.

    This object MUST contain only the properties defined below.
    """

    method: Literal["GET", "POST"] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The HTTP method to use when establishing the connection. Its value "
            "MUST be either GET or POST."
        ),
    )
    query: Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A Schema object containing the definitions for each query parameter. This "
            "schema MUST be of type object and have a properties key."
        ),
    )
    headers: Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A Schema object containing the definitions of the HTTP headers to use "
            "when establishing the connection. This schema MUST be of type object and "
            "have a properties key."
        ),
    )
    binding_version: str = Field(
        default="0.1.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )


class WebSocketsOperationBindings(NonExtendableBaseModel):
    """
    WebSockets Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class WebSocketsMessageBindings(NonExtendableBaseModel):
    """
    WebSockets Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""Solace bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "SolaceChannelBindings",
    "SolaceDestination",
    "SolaceMessageBindings",
    "SolaceOperationBindings",
    "SolaceQueue",
    "SolaceServerBindings",
    "SolaceTopic",
]


from typing import Any, Literal

from pydantic import Field, field_validator, model_validator

from asyncapi3.models.base import Reference
from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null
from asyncapi3.models.schema import Schema


class SolaceServerBindings(NonExtendableBaseModel):
    """
    Solace Server Binding Object.

    This object contains information about the server representation in Solace.
    """

    binding_version: str = Field(
        default="0.4.0",
        alias="bindingVersion",
        description="The current version is 0.4.0.",
    )
    msg_vpn: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="msgVpn",
        description="The Virtual Private Network name on the Solace broker.",
    )
    client_name: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="clientName",
        max_length=160,
        description=(
            "A unique client name to use to register to the appliance. If specified, "
            "it must be a valid Topic name, and a maximum of 160 bytes in length when "
            "encoded as UTF-8."
        ),
    )


class SolaceChannelBindings(NonExtendableBaseModel):
    """
    Solace Channel Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class SolaceQueue(NonExtendableBaseModel):
    """
    Solace Queue.

    Queue configuration for Solace destination.
    """

    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The name of the queue, only applicable when destinationType is 'queue'."
        ),
    )
    topic_subscriptions: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="topicSubscriptions",
        description=(
            "A list of topics that the queue subscribes to, only applicable when "
            "destinationType is 'queue'. If none is given, the queue subscribes to "
            "the topic as represented by the channel name."
        ),
    )
    access_type: Literal["exclusive", "nonexclusive"] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="accessType",
        description=(
            "'exclusive' or 'nonexclusive'. Only applicable when destinationType is "
            "'queue'."
        ),
    )
    max_msg_spool_size: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="maxMsgSpoolSize",
        description=(
            "The maximum amount of message spool that the given queue may use. Only "
            "applicable when destinationType is 'queue'."
        ),
    )
    max_ttl: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="maxTtl",
        description=(
            "The maximum TTL to apply to messages to be spooled. Only applicable when "
            "destinationType is 'queue'."
        ),
    )


class SolaceTopic(NonExtendableBaseModel):
    """
    Solace Topic.

    Topic configuration for Solace destination.
    """

    topic_subscriptions: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="topicSubscriptions",
        description=(
            "A list of topics that the client subscribes to, only applicable when "
            "destinationType is 'topic'. If none is given, the client subscribes to "
            "the topic as represented by the channel name."
        ),
    )


class SolaceDestination(NonExtendableBaseModel):
    """
    Solace Destination.

    Destination Objects are described here.
    """

    binding_version: str = Field(
        default="0.4.0",
        alias="bindingVersion",
        description="The current version is 0.4.0.",
    )
    destination_type: Literal["queue", "topic"] = Field(
        alias="destinationType",
        description=(
            "'queue' or 'topic'. If the type is queue, then the subscriber can bind "
            "to the queue, which in turn will subscribe to the topic as represented "
            "by the channel name or to the provided topicSubscriptions."
        ),
    )
    delivery_mode: Literal["direct", "persistent"] = Field(
        default="persistent",
        alias="deliveryMode",
        description=(
            "'direct' or 'persistent'. This determines the quality of service for "
            "publishing messages. Default is 'persistent'."
        ),
    )
    queue: SolaceQueue | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Queue configuration, only applicable when destinationType is 'queue'."
        ),
    )
    topic: SolaceTopic | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Topic configuration, only applicable when destinationType is 'topic'."
        ),
    )

    @model_validator(mode="after")
    def validate_destination_fields(self) -> "SolaceDestination":
        """Validate that queue/topic fields match destination_type."""
        if self.destination_type == "queue":
            if self.queue is None:
                raise ValueError("queue required for destinationType 'queue'")
            if self.topic is not None:
                raise ValueError("topic not allowed for destinationType 'queue'")
        elif self.destination_type == "topic":
            if self.topic is None:
                raise ValueError("topic required for destinationType 'topic'")
            if self.queue is not None:
                raise ValueError("queue not allowed for destinationType 'topic'")
        return self


class SolaceOperationBindings(NonExtendableBaseModel):
    """
    Solace Operation Binding Object.

    We need the ability to support several bindings for each operation.
    """

    binding_version: str = Field(
        default="0.4.0",
        alias="bindingVersion",
        description="The current version is 0.4.0.",
    )
    destinations: list[SolaceDestination] | None = Field(
        default=None,
        exclude_if=is_null,
        description="Destination Objects are described next.",
    )
    time_to_live: int | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="timeToLive",
        description=(
            "Interval in milliseconds or a Schema Object containing the definition of "
            "the lifetime of the message."
        ),
    )
    priority: int | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The valid priority value range is 0-255 with 0 as the lowest priority and "
            "255 as the highest or a Schema Object containing the definition of the "
            "priority."
        ),
    )
    dmq_eligible: bool | None = Field(
        default=None,
        exclude_if=is_null,
        alias="dmqEligible",
        description=(
            "Set the message to be eligible to be moved to a Dead Message Queue. The "
            "default value is false."
        ),
    )

    @field_validator("priority")
    @classmethod
    def validate_priority(cls, priority: Any) -> Any:
        """Validate priority range when it's an integer."""
        if isinstance(priority, int) and (priority < 0 or priority > 255):
            raise ValueError("priority must be between 0 and 255")
        return priority


class SolaceMessageBindings(NonExtendableBaseModel):
    """
    Solace Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""Base models for AsyncAPI 3.0 specification."""

__all__ = ["ExternalDocumentation", "Reference", "Tag", "Tags"]

from pydantic import AnyUrl, Field

from asyncapi3.models.base_models import ExtendableBaseModel, NonExtendableBaseModel
from asyncapi3.models.helpers import is_null


class Reference(NonExtendableBaseModel):
    """
    Reference Object.

    A simple object to allow referencing other components in the specification,
    internally and externally.

    The Reference Object is defined by JSON Reference and follows the same structure,
    behavior and rules. A JSON Reference SHALL only be used to refer to a schema that
    is formatted in either JSON or YAML. In the case of a YAML-formatted Schema, the
    JSON Reference SHALL be applied to the JSON representation of that schema. The
    JSON representation SHALL be made by applying the conversion described in the Format
    section.

    For this specification, reference resolution is done as defined by the JSON
    Reference specification and not by the JSON Schema specification.
    """

    ref: str = Field(
        alias="$ref",
        description="The reference string.",
    )


class ExternalDocumentation(ExtendableBaseModel):
    """
    External Documentation Object.

    Allows referencing an external resource for extended documentation.
    """

    url: AnyUrl = Field(
        description=(
            "The URL for the target documentation. This MUST be in the form of an "
            "absolute URL."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A short description of the target documentation. CommonMark syntax can "
            "be used for rich text representation."
        ),
    )


class Tag(ExtendableBaseModel):
    """
    Tag Object.

    Allows adding meta data to a single tag.
    """

    name: str = Field(
        description="The name of the tag.",
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A short description for the tag. CommonMark syntax can be used for rich "
            "text representation."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation for this tag.",
    )


# Tags is a type alias for a list of Tag objects
Tags = list[Tag | Reference]



"""Base model classes for AsyncAPI 3.0 specification."""

__all__ = ["ExtendableBaseModel", "NonExtendableBaseModel", "PatternedRootModel"]

import re

from collections.abc import Iterator
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict, RootModel, model_validator

T = TypeVar("T")


class ExtendableBaseModel(BaseModel):
    """
    Base model that allows specification extensions.

    Extensions are fields prefixed with "x-" that follow the pattern
    ^x-[\\w\\d\\.\\x2d_]+$. This model allows extra fields and validates that any
    additional fields match the extension pattern.
    """

    model_config = ConfigDict(
        extra="allow",
        revalidate_instances="always",
        validate_assignment=True,
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )

    @model_validator(mode="after")
    def validate_extensions(self) -> "ExtendableBaseModel":
        """
        Validate that any extra fields match the specification extension pattern.

        Extensions must start with "x-" and follow the regex pattern
        ^x-[\\w\\d\\.\\x2d_]+$.
        """
        if not self.model_extra:
            return self

        extension_pattern = re.compile(r"^x-[\w\d\.\x2d_]+$")
        for key in self.model_extra:
            if not extension_pattern.match(key):
                raise ValueError(
                    f"Field '{key}' does not match specification extension pattern. "
                    f"Extensions must start with 'x-' and contain only word "
                    f"characters, digits, dots, hyphens, and underscores."
                )
        return self


class NonExtendableBaseModel(BaseModel):
    """
    Base model that does not allow specification extensions or extra fields.

    This model forbids any extra fields beyond those explicitly defined.
    """

    model_config = ConfigDict(
        extra="forbid",
        revalidate_instances="always",
        validate_assignment=True,
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )


class PatternedRootModel(RootModel[dict[str, T]], Generic[T]):
    """
    Base class for AsyncAPI patterned objects that validate key patterns.

    This model validates that all keys match the AsyncAPI patterned object key pattern
    ^[A-Za-z0-9\\.\\-_]+$.
    """

    model_config = ConfigDict(
        revalidate_instances="always",
        validate_assignment=True,
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )

    def __iter__(self) -> Iterator[str]:  # type: ignore[override]
        return iter(self.root)

    def __getitem__(self, item: str) -> T:
        return self.root[item]

    def __getattr__(self, item: str) -> T:
        return self.root[item]

    @model_validator(mode="after")
    def validate_patterned_keys(self) -> "PatternedRootModel[T]":
        """
        Validate that all keys in the input data match the AsyncAPI patterned
        object key pattern.

        Keys must match the regex pattern ^[A-Za-z0-9\\.\\-_]+$
        """
        if not self.root:
            return self

        extension_pattern = re.compile(r"^[A-Za-z0-9\.\-_]+$")
        for field_name in self.root:
            if not extension_pattern.match(field_name):
                raise ValueError(
                    f"Field '{field_name}' does not match patterned object key pattern."
                    " Keys must match [A-Za-z0-9\\.\\-_]+"
                )
        return self



"""Operation models for AsyncAPI 3.0 specification."""

__all__ = [
    "Operation",
    "OperationReply",
    "OperationReplyAddress",
    "OperationTrait",
    "Operations",
]

from typing import Literal

from pydantic import Field

from asyncapi3.models.base import ExternalDocumentation, Reference, Tags
from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.bindings import OperationBindingsObject
from asyncapi3.models.helpers import is_null
from asyncapi3.models.security import SecurityScheme


class OperationReplyAddress(ExtendableBaseModel):
    """
    Operation Reply Address Object.

    An object that specifies where an operation has to send the reply.

    For specifying and computing the location of a reply address, a runtime expression
    is used.

    This object MAY be extended with Specification Extensions.
    """

    location: str = Field(
        description=(
            "REQUIRED. A runtime expression that specifies the location of the reply "
            "address."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An optional description of the address. CommonMark syntax can be used "
            "for rich text representation."
        ),
    )


class OperationReply(ExtendableBaseModel):
    """
    Operation Reply Object.

    Describes the reply part that MAY be applied to an Operation Object. If an
    operation implements the request/reply pattern, the reply object represents the
    response message.

    This object MAY be extended with Specification Extensions.
    """

    address: OperationReplyAddress | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Definition of the address that implementations MUST use for the reply."
        ),
    )
    # TODO: How to deal with deref recommendation?
    channel: Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A $ref pointer to the definition of the channel in which this "
            "operation is performed. When address is specified, the address property "
            "of the channel referenced by this property MUST be either null or not "
            "defined. If the operation reply is located inside a root Operation "
            "Object, it MUST point to a channel definition located in the root "
            "Channels Object, and MUST NOT point to a channel definition located in "
            "the Components Object or anywhere else. If the operation reply is "
            "located inside an Operation Object in the Components Object or in the "
            "Replies Object in the Components Object, it MAY point to a Channel "
            "Object in any location. Please note the channel property value MUST be "
            "a Reference Object and, therefore, MUST NOT contain a Channel Object. "
            "However, it is RECOMMENDED that parsers (or other software) dereference "
            "this property for a better development experience."
        ),
    )
    # TODO: How to deal with deref recommendation?
    messages: list[Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of $ref pointers pointing to the supported Message Objects that "
            "can be processed by this operation as reply. It MUST contain a subset of "
            "the messages defined in the channel referenced in this operation reply, "
            "and MUST NOT point to a subset of message definitions located in the "
            "Components Object or anywhere else. Every message processed by this "
            "operation MUST be valid against one, and only one, of the message "
            "objects referenced in this list. Please note the messages property "
            "value MUST be a list of Reference Objects and, therefore, MUST NOT "
            "contain Message Objects. However, it is RECOMMENDED that parsers (or "
            "other software) dereference this property for a better development "
            "experience."
        ),
    )


class OperationTrait(ExtendableBaseModel):
    """
    Operation Trait Object.

    Describes a trait that MAY be applied to an Operation Object. This object MAY
    contain any property from the Operation Object, except the action, channel,
    messages and traits ones.

    If you're looking to apply traits to a message, see the Message Trait Object.

    This object MAY be extended with Specification Extensions.
    """

    title: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A human-friendly title for the operation.",
    )
    summary: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A short summary of what the operation is about.",
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A verbose explanation of the operation. CommonMark syntax can be used "
            "for rich text representation."
        ),
    )
    security: list[SecurityScheme | Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A declaration of which security schemes are associated with this "
            "operation. Only one of the security scheme objects MUST be satisfied to "
            "authorize an operation. In cases where Server Security also applies, "
            "it MUST also be satisfied."
        ),
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of tags for logical grouping and categorization of operations."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation for this operation.",
    )
    bindings: OperationBindingsObject | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map where the keys describe the name of the protocol and the values "
            "describe protocol-specific definitions for the operation."
        ),
    )


class Operation(ExtendableBaseModel):
    """
    Operation Object.

    Describes a specific operation.

    This object MAY be extended with Specification Extensions.
    """

    action: Literal["send", "receive"] = Field(
        description=(
            "Use send when it's expected that the application will send a message "
            "to the given channel, and receive when the application should expect "
            "receiving messages from the given channel."
        ),
    )
    # TODO: How to deal with deref recommendation?
    channel: Reference = Field(
        description=(
            "REQUIRED. A $ref pointer to the definition of the channel in which this "
            "operation is performed. If the operation is located in the root "
            "Operations Object, it MUST point to a channel definition located in the "
            "root Channels Object, and MUST NOT point to a channel definition located "
            "in the Components Object or anywhere else. If the operation is located "
            "in the Components Object, it MAY point to a Channel Object in any "
            "location. Please note the channel property value MUST be a Reference "
            "Object and, therefore, MUST NOT contain a Channel Object. However, "
            "it is RECOMMENDED that parsers (or other software) dereference this "
            "property for a better development experience."
        ),
    )
    title: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A human-friendly title for the operation.",
    )
    summary: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A short summary of what the operation is about.",
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A verbose explanation of the operation. CommonMark syntax can be used "
            "for rich text representation."
        ),
    )
    security: list[SecurityScheme | Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A declaration of which security schemes are associated with this "
            "operation. Only one of the security scheme objects MUST be satisfied to "
            "authorize an operation. In cases where Server Security also applies, "
            "it MUST also be satisfied."
        ),
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of tags for logical grouping and categorization of operations."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation for this operation.",
    )
    bindings: OperationBindingsObject | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map where the keys describe the name of the protocol and the values "
            "describe protocol-specific definitions for the operation."
        ),
    )
    # TODO: Traits MUST be merged using traits merge mechanism?
    traits: list[OperationTrait | Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of traits to apply to the operation object. Traits MUST be merged "
            "using traits merge mechanism. The resulting object MUST be a valid "
            "Operation Object."
        ),
    )
    # TODO: How to deal with deref recommendation?
    messages: list[Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of $ref pointers pointing to the supported Message Objects that "
            "can be processed by this operation. It MUST contain a subset of the "
            "messages defined in the channel referenced in this operation, and "
            "MUST NOT point to a subset of message definitions located in the "
            "Messages Object in the Components Object or anywhere else. Every message "
            "processed by this operation MUST be valid against one, and only one, of "
            "the message objects referenced in this list. Please note the messages "
            "property value MUST be a list of Reference Objects and, therefore, "
            "MUST NOT contain Message Objects. However, it is RECOMMENDED that parsers "
            "(or other software) dereference this property for a better development "
            "experience. Note: excluding this property from the Operation implies "
            "that all messages from the channel will be included. Explicitly set the "
            "messages property to [] if this operation should contain no messages."
        ),
    )
    reply: OperationReply | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description="The definition of the reply in a request-reply operation.",
    )


# Operations is a type alias for a dictionary of Operation objects
Operations = dict[str, Operation | Reference]



"""MQTT bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "MQTTChannelBindings",
    "MQTTLastWill",
    "MQTTMessageBindings",
    "MQTTOperationBindings",
    "MQTTServerBindings",
]

from typing import Literal

from pydantic import Field

from asyncapi3.models.base import Reference
from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null
from asyncapi3.models.schema import Schema


class MQTTLastWill(NonExtendableBaseModel):
    """
    MQTT Last Will and Testament configuration.

    Last Will and Testament configuration. topic, qos, message and retain are
    properties of this object.
    """

    topic: str = Field(
        description="The topic where the Last Will and Testament message will be sent.",
    )
    qos: Literal[0, 1, 2] = Field(
        description=(
            "Defines how hard the broker/client will try to ensure that the Last Will "
            "and Testament message is received. Its value MUST be either 0, 1 or 2."
        ),
    )
    message: str = Field(
        description="Last Will message.",
    )
    retain: bool = Field(
        description=(
            "Whether the broker should retain the Last Will and Testament message or "
            "not."
        ),
    )


class MQTTServerBindings(NonExtendableBaseModel):
    """
    MQTT Server Binding Object.

    This object contains information about the server representation in MQTT.

    This object MUST contain only the properties defined below.
    """

    client_id: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="clientId",
        description="The client identifier.",
    )
    clean_session: bool | None = Field(
        default=None,
        exclude_if=is_null,
        alias="cleanSession",
        description=(
            "Whether to create a persistent connection or not. When false, the "
            "connection will be persistent. This is called clean start in MQTTv5."
        ),
    )
    last_will: MQTTLastWill | None = Field(
        default=None,
        exclude_if=is_null,
        alias="lastWill",
        description="Last Will and Testament configuration.",
    )
    keep_alive: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="keepAlive",
        description=(
            "Interval in seconds of the longest period of time the broker and the "
            "client can endure without sending a message."
        ),
    )
    session_expiry_interval: int | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="sessionExpiryInterval",
        description=(
            "Interval in seconds or a Schema Object containing the definition of the "
            "interval. The broker maintains a session for a disconnected client until "
            "this interval expires."
        ),
    )
    maximum_packet_size: int | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="maximumPacketSize",
        description=(
            "Number of bytes or a Schema Object representing the maximum packet size "
            "the client is willing to accept."
        ),
    )
    binding_version: str = Field(
        default="0.2.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )


class MQTTChannelBindings(NonExtendableBaseModel):
    """
    MQTT Channel Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class MQTTOperationBindings(NonExtendableBaseModel):
    """
    MQTT Operation Binding Object.

    This object contains information about the operation representation in MQTT.

    This object MUST contain only the properties defined below.
    """

    qos: Literal[0, 1, 2] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Defines the Quality of Service (QoS) levels for the message flow between "
            "client and server. Its value MUST be either 0 (At most once delivery), 1 "
            "(At least once delivery), or 2 (Exactly once delivery)."
        ),
    )
    retain: bool | None = Field(
        default=None,
        exclude_if=is_null,
        description="Whether the broker should retain the message or not.",
    )
    message_expiry_interval: int | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="messageExpiryInterval",
        description=(
            "Interval in seconds or a Schema Object containing the definition of the "
            "lifetime of the message."
        ),
    )
    binding_version: str = Field(
        default="0.2.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )


class MQTTMessageBindings(NonExtendableBaseModel):
    """
    MQTT Message Binding Object.

    This object contains information about the message representation in MQTT.

    This object MUST contain only the properties defined below.
    """

    payload_format_indicator: Literal[0, 1] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="payloadFormatIndicator",
        description=(
            "Either: 0 (zero): Indicates that the payload is unspecified bytes, or 1: "
            "Indicates that the payload is UTF-8 encoded character data."
        ),
    )
    correlation_data: Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="correlationData",
        description=(
            "Correlation Data is used by the sender of the request message to identify "
            "which request the response message is for when it is received."
        ),
    )
    content_type: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="contentType",
        description=(
            "String describing the content type of the message payload. This should "
            "not conflict with the contentType field of the associated AsyncAPI "
            "Message object."
        ),
    )
    response_topic: str | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="responseTopic",
        description="The topic (channel URI) for a response message.",
    )
    binding_version: str = Field(
        default="0.2.0",
        alias="bindingVersion",
        description="The version of this binding. If omitted, 'latest' MUST be assumed",
    )



"""Google Cloud Pub/Sub bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "GooglePubSubChannelBindings",
    "GooglePubSubMessageBindings",
    "GooglePubSubMessageStoragePolicy",
    "GooglePubSubOperationBindings",
    "GooglePubSubSchemaDefinition",
    "GooglePubSubSchemaSettings",
    "GooglePubSubServerBindings",
]

from typing import Any, Literal

from pydantic import Field

from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null


class GooglePubSubServerBindings(NonExtendableBaseModel):
    """
    Google Cloud Pub/Sub Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class GooglePubSubMessageStoragePolicy(NonExtendableBaseModel):
    """
    Google Cloud Pub/Sub Message Storage Policy Object.

    Policy constraining the set of Google Cloud Platform regions where messages
    published to the topic may be stored.
    """

    allowed_persistence_regions: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="allowedPersistenceRegions",
        description=(
            "A list of IDs of GCP regions where messages that are published to the "
            "topic may be persisted in storage."
        ),
    )


class GooglePubSubSchemaSettings(NonExtendableBaseModel):
    """
    Google Cloud Pub/Sub Schema Settings Object.

    Settings for validating messages published against a schema.
    """

    encoding: (
        Literal[
            "ENCODING_UNSPECIFIED",
            "JSON",
            "BINARY",
        ]
        | None
    ) = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The encoding of the message. Must be one of the possible Encoding values."
        ),
    )
    first_revision_id: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="firstRevisionId",
        description=(
            "The minimum (inclusive) revision allowed for validating messages."
        ),
    )
    last_revision_id: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="lastRevisionId",
        description=(
            "The maximum (inclusive) revision allowed for validating messages."
        ),
    )
    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The name of the schema that messages published should be validated "
            "against. The format is projects/{project}/schemas/{schema}."
        ),
    )


class GooglePubSubChannelBindings(NonExtendableBaseModel):
    """
    Google Cloud Pub/Sub Channel Binding Object.

    The Channel Bindings Object is used to describe the Google Cloud Pub/Sub specific
    Topic details with AsyncAPI.
    """

    binding_version: str = Field(
        default="0.2.0",
        alias="bindingVersion",
        description="The current version is 0.2.0.",
    )
    labels: dict[str, Any] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An object of key-value pairs. These are used to categorize Cloud "
            "Resources like Cloud Pub/Sub Topics."
        ),
    )
    message_retention_duration: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="messageRetentionDuration",
        description=(
            "Indicates the minimum duration to retain a message after it is published "
            "to the topic. Must be a valid Duration."
        ),
    )
    message_storage_policy: GooglePubSubMessageStoragePolicy | None = Field(
        default=None,
        exclude_if=is_null,
        alias="messageStoragePolicy",
        description=(
            "Policy constraining the set of Google Cloud Platform regions where "
            "messages published to the topic may be stored."
        ),
    )
    schema_settings: GooglePubSubSchemaSettings | None = Field(
        default=None,
        exclude_if=is_null,
        alias="schemaSettings",
        description="Settings for validating messages published against a schema.",
    )


class GooglePubSubOperationBindings(NonExtendableBaseModel):
    """
    Google Cloud Pub/Sub Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class GooglePubSubSchemaDefinition(NonExtendableBaseModel):
    """
    Google Cloud Pub/Sub Schema Definition Object.

    Describes the schema used to validate the payload of this message.
    """

    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="The name of the schema.",
    )


class GooglePubSubMessageBindings(NonExtendableBaseModel):
    """
    Google Cloud Pub/Sub Message Binding Object.

    The Message Binding Object is used to describe the Google Cloud Pub/Sub specific
    PubsubMessage details, alongside with pertinent parts of the Google Cloud Pub/Sub
    Schema Object, with AsyncAPI.
    """

    binding_version: str = Field(
        default="0.2.0",
        alias="bindingVersion",
        description="The current version is 0.2.0.",
    )
    attributes: dict[str, Any] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Attributes for this message. If this field is empty, the message must "
            "contain non-empty data. This can be used to filter messages on the "
            "subscription."
        ),
    )
    ordering_key: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="orderingKey",
        description=(
            "If non-empty, identifies related messages for which publish order should "
            "be respected. For more information, see ordering messages."
        ),
    )
    schema: GooglePubSubSchemaDefinition | None = Field(  # type: ignore[assignment]
        default=None,
        exclude_if=is_null,
        description=(
            "Describes the schema used to validate the payload of this message."
        ),
    )



"""Message models for AsyncAPI 3.0 specification."""

__all__ = [
    "Message",
    "MessageExample",
    "MessageTrait",
    "Messages",
]

from typing import Any

from pydantic import Field, model_validator

from asyncapi3.models.base import ExternalDocumentation, Reference, Tags
from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.bindings import MessageBindingsObject
from asyncapi3.models.helpers import is_null
from asyncapi3.models.schema import MultiFormatSchema, Schema
from asyncapi3.models.security import CorrelationID


class MessageExample(ExtendableBaseModel):
    """
    Message Example Object.

    Message Example Object represents an example of a Message Object and MUST contain
    either headers and/or payload fields.

    This object MAY be extended with Specification Extensions.
    """

    headers: dict[str, Any] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The value of this field MUST validate against the Message Object's "
            "headers field."
        ),
    )
    payload: Any | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The value of this field MUST validate against the Message Object's "
            "payload field."
        ),
    )
    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A machine-friendly name.",
    )
    summary: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A short summary of what the example is about.",
    )

    @model_validator(mode="after")
    def validate_headers_or_payload(self) -> "MessageExample":
        """
        Validate that MessageExample contains either headers and/or payload fields.

        According to AsyncAPI specification: "Message Example Object represents an
        example of a Message Object and MUST contain either headers and/or payload
        fields."
        """
        if self.headers is None and self.payload is None:
            raise ValueError(
                "MessageExample MUST contain either headers and/or payload fields"
            )
        return self


class MessageTrait(ExtendableBaseModel):
    """
    Message Trait Object.

    Describes a trait that MAY be applied to a Message Object. This object MAY contain
    any property from the Message Object, except payload and traits.

    If you're looking to apply traits to an operation, see the Operation Trait Object.

    This object MAY be extended with Specification Extensions.
    """

    headers: MultiFormatSchema | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Schema definition of the application headers. Schema MUST be a map of "
            "key-value pairs. It MUST NOT define the protocol headers. If this is a "
            "Schema Object, then the schemaFormat will be assumed to be "
            "'application/vnd.aai.asyncapi+json;version=asyncapi' where the version "
            "is equal to the AsyncAPI Version String."
        ),
    )
    correlation_id: CorrelationID | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="correlationId",
        description=(
            "Definition of the correlation ID used for message tracing or matching."
        ),
    )
    content_type: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="contentType",
        description=(
            "The content type to use when encoding/decoding a message's payload. "
            "The value MUST be a specific media type (e.g. application/json). "
            "When omitted, the value MUST be the one specified on the "
            "defaultContentType field."
        ),
    )
    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A machine-friendly name for the message.",
    )
    title: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A human-friendly title for the message.",
    )
    summary: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A short summary of what the message is about.",
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A verbose explanation of the message. CommonMark syntax can be used for "
            "rich text representation."
        ),
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of tags for logical grouping and categorization of messages."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation for this message.",
    )
    bindings: MessageBindingsObject | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map where the keys describe the name of the protocol and the values "
            "describe protocol-specific definitions for the message."
        ),
    )
    examples: list[MessageExample] | None = Field(
        default=None,
        exclude_if=is_null,
        description="List of examples.",
    )


class Message(ExtendableBaseModel):
    """
    Message Object.

    Describes a message received on a given channel and operation.

    This object MAY be extended with Specification Extensions.
    """

    headers: MultiFormatSchema | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Schema definition of the application headers. Schema MUST be a map of "
            "key-value pairs. It MUST NOT define the protocol headers. If this is a "
            "Schema Object, then the schemaFormat will be assumed to be "
            "'application/vnd.aai.asyncapi+json;version=asyncapi' where the version "
            "is equal to the AsyncAPI Version String."
        ),
    )
    payload: MultiFormatSchema | Schema | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Definition of the message payload. If this is a Schema Object, then the "
            "schemaFormat will be assumed to be "
            "'application/vnd.aai.asyncapi+json;version=asyncapi' where the version "
            "is equal to the AsyncAPI Version String."
        ),
    )
    correlation_id: CorrelationID | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="correlationId",
        description=(
            "Definition of the correlation ID used for message tracing or matching."
        ),
    )
    content_type: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="contentType",
        description=(
            "The content type to use when encoding/decoding a message's payload. "
            "The value MUST be a specific media type (e.g. application/json). "
            "When omitted, the value MUST be the one specified on the "
            "defaultContentType field."
        ),
    )
    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A machine-friendly name for the message.",
    )
    title: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A human-friendly title for the message.",
    )
    summary: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A short summary of what the message is about.",
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A verbose explanation of the message. CommonMark syntax can be used for "
            "rich text representation."
        ),
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of tags for logical grouping and categorization of messages."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation for this message.",
    )
    bindings: MessageBindingsObject | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map where the keys describe the name of the protocol and the values "
            "describe protocol-specific definitions for the message."
        ),
    )
    examples: list[MessageExample] | None = Field(
        default=None,
        exclude_if=is_null,
        description="List of examples.",
    )
    # TODO: Traits MUST be merged using traits merge mechanism?
    traits: list[MessageTrait | Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of traits to apply to the message object. Traits MUST be merged "
            "using traits merge mechanism. The resulting object MUST be a valid "
            "Message Object."
        ),
    )


# Messages is a type alias for a dictionary of Message objects
Messages = dict[str, Message | Reference]



"""Security models for AsyncAPI 3.0 specification."""

__all__ = [
    "CorrelationID",
    "OAuthFlow",
    "OAuthFlows",
    "SecurityScheme",
]

from typing import Literal

from pydantic import AnyUrl, Field, field_validator, model_validator

from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.helpers import is_null


class CorrelationID(ExtendableBaseModel):
    """
    Correlation ID Object.

    An object that specifies an identifier at design time that can used for message
    tracing and correlation.

    For specifying and computing the location of a Correlation ID, a runtime
    expression is used.

    This object MAY be extended with Specification Extensions.
    """

    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An optional description of the identifier. CommonMark syntax can be used "
            "for rich text representation."
        ),
    )
    location: str = Field(
        description=(
            "REQUIRED. A runtime expression that specifies the location of the "
            "correlation ID."
        ),
    )

    @field_validator("location")
    @classmethod
    def validate_runtime_expression(cls, location: str) -> str:
        """
        Validate that location contains a valid runtime expression.

        Runtime expressions must start with '$message.' according to
        AsyncAPI 3.0 specification.
        """
        if not location.startswith("$message."):
            raise ValueError(
                "location must be a runtime expression starting with '$message.'"
            )
        return location


class OAuthFlow(ExtendableBaseModel):
    """
    OAuth Flow Object.

    Configuration details for a supported OAuth Flow.

    This object MAY be extended with Specification Extensions.
    """

    authorization_url: AnyUrl | None = Field(
        default=None,
        exclude_if=is_null,
        alias="authorizationUrl",
        description=(
            "REQUIRED for: oauth2 ('implicit', 'authorizationCode'). The "
            "authorization URL to be used for this flow. This MUST be in the form "
            "of an absolute URL."
        ),
    )
    token_url: AnyUrl | None = Field(
        default=None,
        exclude_if=is_null,
        alias="tokenUrl",
        description=(
            "REQUIRED for: oauth2 ('password', 'clientCredentials', "
            "'authorizationCode'). The token URL to be used for this flow. This MUST "
            "be in the form of an absolute URL."
        ),
    )
    refresh_url: AnyUrl | None = Field(
        default=None,
        exclude_if=is_null,
        alias="refreshUrl",
        description=(
            "Applied to oauth2 type. The URL to be used for obtaining refresh tokens. "
            "This MUST be in the form of an absolute URL."
        ),
    )
    available_scopes: dict[str, str] = Field(
        alias="availableScopes",
        description=(
            "REQUIRED for oauth2 type. The available scopes for the OAuth2 security "
            "scheme. A map between the scope name and a short description for it."
        ),
    )


class OAuthFlows(ExtendableBaseModel):
    """
    OAuth Flows Object.

    Allows configuration of the supported OAuth Flows.

    This object MAY be extended with Specification Extensions.
    """

    implicit: OAuthFlow | None = Field(
        default=None,
        exclude_if=is_null,
        description="Configuration for the OAuth Implicit flow.",
    )
    password: OAuthFlow | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Configuration for the OAuth Resource Owner Protected Credentials flow."
        ),
    )
    client_credentials: OAuthFlow | None = Field(
        default=None,
        exclude_if=is_null,
        alias="clientCredentials",
        description="Configuration for the OAuth Client Credentials flow.",
    )
    authorization_code: OAuthFlow | None = Field(
        default=None,
        exclude_if=is_null,
        alias="authorizationCode",
        description="Configuration for the OAuth Authorization Code flow.",
    )

    @model_validator(mode="after")
    def validate_oauth_flows_requirements(self) -> "OAuthFlows":
        """
        Validate OAuthFlows field requirements based on flow types.

        Performs validation according to AsyncAPI 3.0 specification requirements
        for different OAuth flow types.
        """
        self._validate_implicit_flow()
        self._validate_password_flow()
        self._validate_client_credentials_flow()
        self._validate_authorization_code_flow()
        return self

    def _validate_implicit_flow(self) -> None:
        """Validate implicit flow requirements."""
        if self.implicit is not None and self.implicit.authorization_url is None:
            raise ValueError("authorizationUrl is required for implicit flow")

    def _validate_password_flow(self) -> None:
        """Validate password flow requirements."""
        if self.password is not None and self.password.token_url is None:
            raise ValueError("tokenUrl is required for password flow")

    def _validate_client_credentials_flow(self) -> None:
        """Validate clientCredentials flow requirements."""
        if (
            self.client_credentials is not None
            and self.client_credentials.token_url is None
        ):
            raise ValueError("tokenUrl is required for clientCredentials flow")

    def _validate_authorization_code_flow(self) -> None:
        """Validate authorizationCode flow requirements."""
        if self.authorization_code is not None:
            if self.authorization_code.authorization_url is None:
                raise ValueError(
                    "authorizationUrl is required for authorizationCode flow"
                )
            if self.authorization_code.token_url is None:
                raise ValueError("tokenUrl is required for authorizationCode flow")


class SecurityScheme(ExtendableBaseModel):
    """
    Security Scheme Object.

    Defines a security scheme that can be used by the operations. Supported schemes
    are:

    - User/Password.
    - API key (either as user or as password).
    - X.509 certificate.
    - End-to-end encryption (either symmetric or asymmetric).
    - HTTP authentication.
    - HTTP API key.
    - OAuth2's common flows (Implicit, Resource Owner Protected Credentials, Client
      Credentials and Authorization Code) as defined in RFC6749.
    - OpenID Connect Discovery.
    - SASL (Simple Authentication and Security Layer) as defined in RFC4422.
    """

    type_: Literal[
        "apiKey",
        "asymmetricEncryption",
        "gssapi",
        "http",
        "httpApiKey",
        "oauth2",
        "openIdConnect",
        "plain",
        "scramSha256",
        "scramSha512",
        "symmetricEncryption",
        "userPassword",
        "X509",
    ] = Field(
        alias="type",
        description=(
            "REQUIRED for any type. The type of the security scheme. Valid values "
            "are 'userPassword', 'apiKey', 'X509', 'symmetricEncryption', "
            "'asymmetricEncryption', 'httpApiKey', 'http', 'oauth2', 'openIdConnect', "
            "'plain', 'scramSha256', 'scramSha512', and 'gssapi'."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A short description for security type. CommonMark syntax MAY be used "
            "for rich text representation."
        ),
    )
    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "REQUIRED for httpApiKey scheme. The name of the header, query or cookie "
            "parameter to be used."
        ),
    )
    in_: (
        Literal[
            "user",
            "password",
            "query",
            "header",
            "cookie",
        ]
        | None
    ) = Field(
        default=None,
        exclude_if=is_null,
        alias="in",
        description=(
            "REQUIRED for apiKey or httpApiKey type. The location of the API key. "
            "Valid values are 'user' and 'password' for apiKey and 'query', 'header' "
            "or 'cookie' for httpApiKey."
        ),
    )
    scheme: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "REQUIRED for http type. The name of the HTTP Authorization scheme to be "
            "used in the Authorization header as defined in RFC7235."
        ),
    )
    bearer_format: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="bearerFormat",
        description=(
            "Used with http ('bearer') type. A hint to the client to identify how the "
            "bearer token is formatted. Bearer tokens are usually generated by an "
            "authorization server, so this information is primarily for documentation "
            "purposes."
        ),
    )
    flows: OAuthFlows | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "REQUIRED for oauth2 type. An object containing configuration information "
            "for the flow types supported."
        ),
    )
    open_id_connect_url: AnyUrl | None = Field(
        default=None,
        exclude_if=is_null,
        alias="openIdConnectUrl",
        description=(
            "REQUIRED for openIdConnect type. OpenId Connect URL to discover OAuth2 "
            "configuration values. This MUST be in the form of an absolute URL."
        ),
    )
    scopes: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Used with oauth2 or openIdConnect type. List of the needed scope names. "
            "An empty array means no scopes are needed."
        ),
    )

    @model_validator(mode="after")
    def validate_security_scheme_dependencies(self) -> "SecurityScheme":
        """
        Validate SecurityScheme field dependencies based on type.

        Performs validation according to AsyncAPI 3.0 specification requirements
        for different security scheme types.
        """
        self._validate_api_key_requirements()
        self._validate_bearer_format_requirements()
        self._validate_http_api_key_requirements()
        self._validate_http_requirements()
        self._validate_oauth2_requirements()
        self._validate_openid_connect_requirements()
        self._validate_scopes_requirements()

        return self

    def _validate_api_key_requirements(self) -> None:
        """Validate apiKey type requirements."""
        if self.type_ == "apiKey":
            if self.in_ is None:
                raise ValueError("in is required for apiKey type")
            if self.in_ not in ("user", "password"):
                raise ValueError("in must be 'user' or 'password' for apiKey type")

    def _validate_http_api_key_requirements(self) -> None:
        """Validate httpApiKey type requirements."""
        if self.type_ == "httpApiKey":
            if self.name is None:
                raise ValueError("name is required for httpApiKey type")
            if self.in_ is None:
                raise ValueError("in is required for httpApiKey type")
            if self.in_ not in ("query", "header", "cookie"):
                raise ValueError(
                    "in must be 'query', 'header', or 'cookie' for httpApiKey type"
                )

    def _validate_http_requirements(self) -> None:
        """Validate http type requirements."""
        if self.type_ == "http" and self.scheme is None:
            raise ValueError("scheme is required for http type")

    def _validate_bearer_format_requirements(self) -> None:
        """Validate bearerFormat requirements."""
        if self.bearer_format is not None and (
            self.type_ != "http" or self.scheme != "bearer"
        ):
            raise ValueError(
                "bearerFormat can only be used with http type and bearer scheme"
            )

    def _validate_oauth2_requirements(self) -> None:
        """Validate oauth2 type requirements."""
        if self.type_ == "oauth2" and self.flows is None:
            raise ValueError("flows is required for oauth2 type")

    def _validate_openid_connect_requirements(self) -> None:
        """Validate openIdConnect type requirements."""
        if self.type_ == "openIdConnect" and self.open_id_connect_url is None:
            raise ValueError("openIdConnectUrl is required for openIdConnect type")

    def _validate_scopes_requirements(self) -> None:
        """Validate scopes requirements."""
        if self.scopes is not None and self.type_ not in ("oauth2", "openIdConnect"):
            raise ValueError(
                "scopes can only be used with oauth2 or openIdConnect type"
            )



"""IBM MQ bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "IBMMQChannelBindings",
    "IBMMQMessageBindings",
    "IBMMQOperationBindings",
    "IBMMQQueue",
    "IBMMQServerBindings",
    "IBMMQTopic",
]

from typing import Literal

from pydantic import Field, field_validator, model_validator

from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null


# TODO: Depends on parent object
class IBMMQServerBindings(NonExtendableBaseModel):
    """
    IBM MQ Server Binding Object.

    This object contains server connection information about the IBM MQ server, referred
    to as an IBM MQ queue manager. This object contains additional connectivity
    information not possible to represent within the core AsyncAPI specification.

    This object MUST contain only the properties defined below.
    """

    # TODO: MUST NOT be specified for URI Scheme http:// or file://
    group_id: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="groupId",
        description=(
            "Defines a logical group of IBM MQ server objects. This is necessary to "
            "specify multi-endpoint configurations used in high availability "
            "deployments. If omitted, the server object is not part of a group. MUST "
            "NOT be specified for URI Scheme http:// or file://."
        ),
    )
    # TODO: MUST NOT be specified for URI Scheme ibmmq://
    ccdt_queue_manager_name: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="ccdtQueueManagerName",
        description=(
            "The name of the IBM MQ queue manager to bind to in the CCDT file. "
            "Optional, defaults to *. MUST NOT be specified for URI Scheme ibmmq://."
        ),
    )
    # TODO: MUST NOT be specified for protocol ibmmq or URI Scheme file:// or http://
    cipher_spec: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="cipherSpec",
        description=(
            "The recommended cipher specification used to establish a TLS connection "
            "between the client and the IBM MQ queue manager. More information on "
            "SSL/TLS cipher specifications supported by IBM MQ can be found in the "
            "IBM MQ Knowledge Center. Optional, defaults to ANY. MUST NOT be specified "
            "for protocol ibmmq or URI Scheme file:// or http://."
        ),
    )
    # TODO: MUST NOT be specified for URI Scheme file:// or http://
    multi_endpoint_server: bool | None = Field(
        default=None,
        exclude_if=is_null,
        alias="multiEndpointServer",
        description=(
            "If multiEndpointServer is true then multiple connections can be workload "
            "balanced and applications should not make assumptions as to where "
            "messages are processed. Where message ordering, or affinity to specific "
            "message resources is necessary, a single endpoint "
            "(multiEndpointServer = false) may be required. Optional, defaults to "
            "false. MUST NOT be specified for URI Scheme file:// or http://."
        ),
    )
    heart_beat_interval: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="heartBeatInterval",
        description=(
            "The recommended value (in seconds) for the heartbeat sent to the queue "
            "manager during periods of inactivity. A value of zero means that no heart "
            "beats are sent. A value of 1 means that the client will use the value "
            "defined by the queue manager. More information on heart beat interval can "
            "be found in the IBM MQ Knowledge Center. Optional, defaults to 300. MUST "
            "be 0-999999."
        ),
    )
    binding_version: str = Field(
        default="0.1.0",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )

    @field_validator("heart_beat_interval")
    @classmethod
    def validate_heart_beat_interval(
        cls, heart_beat_interval: int | None
    ) -> int | None:
        if heart_beat_interval is not None and (
            heart_beat_interval < 0 or heart_beat_interval > 999999
        ):
            raise ValueError("heartBeatInterval MUST be 0-999999")
        return heart_beat_interval


class IBMMQQueue(NonExtendableBaseModel):
    """
    IBM MQ Queue.

    Defines the properties of a queue.
    """

    object_name: str = Field(
        max_length=48,
        alias="objectName",
        description=(
            "Defines the name of the IBM MQ queue associated with the channel. "
            "A value MUST be specified. MUST NOT exceed 48 characters in length. "
            "MUST be a valid IBM MQ queue name."
        ),
    )
    is_partitioned: bool = Field(
        default=False,
        alias="isPartitioned",
        description=(
            "Defines if the queue is a cluster queue and therefore partitioned. If "
            "true, a binding option MAY be specified when accessing the queue. More "
            "information on binding options can be found in the IBM MQ Knowledge "
            "Center. Optional, defaults to false. If false, binding options SHOULD "
            "NOT be specified when accessing the queue."
        ),
    )
    exclusive: bool = Field(
        default=False,
        description=(
            "Specifies if it is recommended to open the queue exclusively. Optional, "
            "defaults to false."
        ),
    )


class IBMMQTopic(NonExtendableBaseModel):
    """
    IBM MQ Topic.

    Defines the properties of a topic.
    """

    # TODO: OPTIONAL Note: if specified, SHALL override AsyncAPI channel name.
    string: str | None = Field(
        default=None,
        exclude_if=is_null,
        max_length=10240,
        description=(
            "The value of the IBM MQ topic string to be used. Optional. Note: if "
            "specified, SHALL override AsyncAPI channel name. MUST NOT exceed 10240 "
            "characters in length. MAY coexist with topic.objectName."
        ),
    )
    # TODO: OPTIONAL Note: if specified, SHALL override AsyncAPI channel name.
    object_name: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="objectName",
        max_length=48,
        description=(
            "The name of the IBM MQ topic object. Optional. Note: if specified, SHALL "
            "override AsyncAPI channel name. MUST NOT exceed 48 characters in length. "
            "MAY coexist with topic.string."
        ),
    )
    durable_permitted: bool = Field(
        default=True,
        alias="durablePermitted",
        description=(
            "Defines if the subscription may be durable. Optional, defaults to true."
        ),
    )
    last_msg_retained: bool = Field(
        default=False,
        alias="lastMsgRetained",
        description=(
            "Defines if the last message published will be made available to new "
            "subscriptions. Optional, defaults to false."
        ),
    )


class IBMMQChannelBindings(NonExtendableBaseModel):
    """
    IBM MQ Channel Binding Object.

    This object contains information about the channel representation in IBM MQ. Each
    channel corresponds to a Queue or Topic within IBM MQ.

    This object MUST contain only the properties defined below.
    """

    destination_type: Literal["topic", "queue"] = Field(
        default="topic",
        alias="destinationType",
        description=(
            "Defines the type of AsyncAPI channel. Optional, defaults to topic. MUST "
            "be either topic or queue. For type topic, the AsyncAPI channel name "
            "MUST be assumed for the IBM MQ topic string unless overridden."
        ),
    )
    queue: IBMMQQueue | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Defines the properties of a queue. REQUIRED if destinationType = queue. "
            "queue and topic fields MUST NOT coexist within a channel binding."
        ),
    )
    topic: IBMMQTopic | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Defines the properties of a topic. OPTIONAL if destinationType = topic. "
            "queue and topic fields MUST NOT coexist within a channel binding."
        ),
    )
    max_msg_length: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="maxMsgLength",
        description=(
            "The maximum length of the physical message (in bytes) accepted by the "
            "Topic or Queue. Messages produced that are greater in size than this "
            "value may fail to be delivered. More information on the maximum message "
            "length can be found in the IBM MQ Knowledge Center. Optional, defaults to "
            "negotiated on IBM MQ channel. MUST be 0-104,857,600 bytes (100 MB)."
        ),
    )
    binding_version: str = Field(
        default="0.1.0",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )

    @field_validator("max_msg_length")
    @classmethod
    def validate_max_msg_length(cls, max_msg_length: int | None) -> int | None:
        if max_msg_length is not None and (
            max_msg_length < 0 or max_msg_length > 104857600
        ):
            raise ValueError("maxMsgLength MUST be 0-104,857,600 bytes (100 MB)")
        return max_msg_length

    @model_validator(mode="after")
    def validate_queue_topic_constraints(self) -> "IBMMQChannelBindings":
        # queue and topic fields MUST NOT coexist within a channel binding
        if self.queue is not None and self.topic is not None:
            raise ValueError(
                "queue and topic fields MUST NOT coexist within a channel binding"
            )

        # queue REQUIRED if destinationType = queue
        if self.destination_type == "queue" and self.queue is None:
            raise ValueError("queue must be provided when destinationType='queue'")

        return self


class IBMMQOperationBindings(NonExtendableBaseModel):
    """
    IBM MQ Operation Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class IBMMQMessageBindings(NonExtendableBaseModel):
    """
    IBM MQ Message Binding Object.

    This object contains information about the message representation in IBM MQ.

    This object MUST contain only the properties defined below.
    """

    type: Literal["string", "jms", "binary"] = Field(
        default="string",
        description="The type of the message.",
    )
    headers: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Defines the IBM MQ message headers to include with this message. More "
            "than one header can be specified as a comma separated list. Supporting "
            "information on IBM MQ message formats can be found in the IBM MQ "
            "Knowledge Center. Optional if type = binary. headers MUST NOT be "
            "specified if type = string or jms."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Provides additional information for application developers: describes "
            "the message type or format."
        ),
    )
    expiry: int | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The recommended setting the client should use for the TTL (Time-To-Live) "
            "of the message. This is a period of time expressed in milliseconds and "
            "set by the application that puts the message. expiry values are API "
            "dependant e.g., MQI and JMS use different units of time and default "
            "values for unlimited. General information on IBM MQ message expiry can "
            "be found in the IBM MQ Knowledge Center. Optional, defaults to "
            "unlimited. expiry value MUST be either zero (unlimited) or greater than "
            "zero."
        ),
    )
    binding_version: str = Field(
        default="0.1.0",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )

    @field_validator("expiry")
    @classmethod
    def validate_expiry(cls, expiry: int | None) -> int | None:
        if expiry is not None and expiry < 0:
            raise ValueError(
                "expiry value MUST be either zero (unlimited) or greater than zero"
            )
        return expiry

    @model_validator(mode="after")
    def validate_headers_constraints(self) -> "IBMMQMessageBindings":
        # headers MUST NOT be specified if type = string or jms
        if self.type in ["string", "jms"] and self.headers is not None:
            raise ValueError(
                "headers MUST NOT be specified if type = 'string' or 'jms'"
            )
        return self



"""SQS bindings models for AsyncAPI 3.0 specification."""

__all__ = [
    "SQSChannelBindings",
    "SQSIdentifier",
    "SQSMessageBindings",
    "SQSOperationBindings",
    "SQSPolicy",
    "SQSQueue",
    "SQSRedrivePolicy",
    "SQSServerBindings",
    "SQSStatement",
]

from typing import Any, Literal

from pydantic import Field

from asyncapi3.models.base_models import NonExtendableBaseModel
from asyncapi3.models.helpers import is_null


class SQSServerBindings(NonExtendableBaseModel):
    """
    SQS Server Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """


class SQSStatement(NonExtendableBaseModel):
    """
    SQS Statement.

    Controls a permission for this queue.
    """

    effect: Literal["Allow", "Deny"] = Field(
        description="Required. Either 'Allow' or 'Deny'.",
    )
    principal: str | dict[str, str | list[str]] = Field(
        description=(
            "Required. The AWS account(s) or resource ARN(s) that the statement "
            "applies to."
        ),
    )
    action: str | list[str] = Field(
        description=(
            "Required. The SQS permission being allowed or denied e.g. sqs:SendMessage."
        ),
    )
    resource: str | list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        description="The resource(s) that this policy applies to.",
    )
    condition: dict[str, Any] | list[dict[str, Any]] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "Specific circumstances under which the policy grants permission."
        ),
    )


class SQSPolicy(NonExtendableBaseModel):
    """
    SQS Policy.

    The security policy for the SQS Queue.
    """

    statements: list[SQSStatement] = Field(
        alias="Statements",
        description=(
            "Required. An array of Statement objects, each of which controls a "
            "permission for this queue."
        ),
    )


class SQSIdentifier(NonExtendableBaseModel):
    """
    SQS Identifier.

    Used to identify an endpoint or queue.
    """

    arn: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The target is an ARN. For example, for SQS, the identifier may be an ARN, "
            "which will be of the form: arn:aws:sqs:{region}:{account-id}:{queueName}."
        ),
    )
    name: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The endpoint is identified by a name, which corresponds to an identifying "
            "field called 'name' of a binding for that protocol on this publish "
            "Operation Object. For example, if the protocol is 'sqs' then the name "
            "refers to the name field sqs binding."
        ),
    )


class SQSRedrivePolicy(NonExtendableBaseModel):
    """
    SQS Redrive Policy.

    Prevent poison pill messages by moving un-processable messages to an SQS dead letter
    queue.
    """

    dead_letter_queue: SQSIdentifier = Field(
        alias="deadLetterQueue",
        description="The SQS queue to use as a dead letter queue (DLQ).",
    )
    max_receive_count: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="maxReceiveCount",
        description=(
            "The number of times a message is delivered to the source queue before "
            "being moved to the dead-letter queue. Default is 10."
        ),
    )


class SQSQueue(NonExtendableBaseModel):
    """
    SQS Queue.

    A definition of the queue that will be used as the channel.
    """

    name: str = Field(
        description=(
            "Required. The name of the queue. When an SNS Operation Binding Object "
            "references an SQS queue by name, the identifier should be the one in "
            "this field."
        ),
    )
    fifo_queue: bool = Field(
        alias="fifoQueue",
        description="Required. Is this a FIFO queue?",
    )
    deduplication_scope: Literal["messageGroup", "queue"] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="deduplicationScope",
        description=(
            "Specifies whether message deduplication occurs at the message group or "
            "queue level. Valid values are messageGroup and queue. This property "
            "applies only to high throughput for FIFO queues."
        ),
    )
    fifo_throughput_limit: Literal["perQueue", "perMessageGroupId"] | None = Field(
        default=None,
        exclude_if=is_null,
        alias="fifoThroughputLimit",
        description=(
            "Specifies whether the FIFO queue throughput quota applies to the entire "
            "queue or per message group. Valid values are perQueue and "
            "perMessageGroupId. The perMessageGroupId value is allowed only when the "
            "value for DeduplicationScope is messageGroup. Setting both these values "
            "as such will enable high throughput on a FIFO queue. As above, this "
            "property applies only to high throughput for FIFO queues."
        ),
    )
    delivery_delay: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="deliveryDelay",
        description=(
            "The number of seconds to delay before a message sent to the queue can be "
            "received. Used to create a delay queue. Range is 0 to 15 minutes. "
            "Defaults to 0."
        ),
    )
    visibility_timeout: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="visibilityTimeout",
        description=(
            "The length of time, in seconds, that a consumer locks a message - hiding "
            "it from reads - before it is unlocked and can be read again. Range from "
            "0 to 12 hours (43200 seconds). Defaults to 30 seconds."
        ),
    )
    receive_message_wait_time: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="receiveMessageWaitTime",
        description=(
            "Determines if the queue uses short polling or long polling. Set to zero "
            "(the default) the queue reads available messages and returns immediately. "
            "Set to a non-zero integer, long polling waits the specified number of "
            "seconds for messages to arrive before returning."
        ),
    )
    message_retention_period: int | None = Field(
        default=None,
        exclude_if=is_null,
        alias="messageRetentionPeriod",
        description=(
            "How long to retain a message on the queue in seconds, unless deleted. The "
            "range is 60 (1 minute) to 1,209,600 (14 days). The default is 345,600 "
            "(4 days)."
        ),
    )
    redrive_policy: SQSRedrivePolicy | None = Field(
        default=None,
        exclude_if=is_null,
        alias="redrivePolicy",
        description=(
            "Prevent poison pill messages by moving un-processable messages to an SQS "
            "dead letter queue."
        ),
    )
    policy: SQSPolicy | None = Field(
        default=None,
        exclude_if=is_null,
        description="The security policy for the SQS Queue.",
    )
    tags: dict[str, Any] | None = Field(
        default=None,
        exclude_if=is_null,
        description="Key-value pairs that represent AWS tags on the queue.",
    )


class SQSChannelBindings(NonExtendableBaseModel):
    """
    SQS Channel Binding Object.

    Use the Channel Binding Operation for Point-to-Point SQS channels.

    There are three likely scenarios for use of the Channel Binding Object:

    - One file defines both publish and subscribe operations, for example if we were
      implementing the work queue pattern to offload work from an HTTP API endpoint to
      a worker process. In this case the channel would be defined on the Channel Object
      in that single file.
    - The producer and consumer both have an AsyncAPI specification file, and the
      producer is raising an event, for example interop between microservices, and the
      producer 'owns' the channel definition and thus has the SQS Binding on its
      Channel Object.
    - The producer and consumer both have an AsyncAPI specification file, and the
      consumer receives commands, for example interop between microservices, and the
      consumer 'owns' the channel definition and thus has the SQS Binding on its
      Channel Object.

    An SQS queue can set up a Dead Letter Queue as part of a Redelivery Policy. To
    support this requirement, the Channel Binding Object allows you to define both a
    Queue Object to use as the Channel or target in a publish Operation and a Dead
    Letter Queue. You can then refer to the Dead letter Queue in the Redrive Policy
    using the Identifier Object and setting the name field to match the name field of
    your Dead Letter Queue Object. (If you define the DLQ externally, the Identifier
    also supports an ARN).
    """

    queue: SQSQueue = Field(
        description=(
            "Required. A definition of the queue that will be used as the channel."
        ),
    )
    dead_letter_queue: SQSQueue | None = Field(
        default=None,
        exclude_if=is_null,
        alias="deadLetterQueue",
        description=(
            "A definition of the queue that will be used for un-processable messages."
        ),
    )
    binding_version: str = Field(
        default="0.3.0",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )


class SQSOperationBindings(NonExtendableBaseModel):
    """
    SQS Operation Binding Object.

    On an Operation Binding Object we support an array of Queue objects. Members of this
    array may be Queue Objects that define the endpoint field required by an SNS
    Operation Binding Object delivering by the SQS protocol or Queue Objects that define
    the Dead Letter Queue used by either the Redrive Policy of the SNS Subscription (see
    the SNS Binding Object) or the Redrive Policy of the SQS Queue. The name of the
    Queue Object is used by an Identifier field on either the endpoint field of the
    SNS Operation Object of deadLetterQueue on the Redrive Policy to identify the
    required member of this array.
    """

    queues: list[SQSQueue] = Field(
        description=(
            "Required. Queue objects that are either the endpoint for an SNS Operation "
            "Binding Object, or the deadLetterQueue of the SQS Operation Binding "
            "Object."
        ),
    )
    binding_version: str = Field(
        default="0.3.0",
        alias="bindingVersion",
        description="The version of this binding. Optional, defaults to latest.",
    )


class SQSMessageBindings(NonExtendableBaseModel):
    """
    SQS Message Binding Object.

    This object MUST NOT contain any properties. Its name is reserved for future use.
    """



"""Server models for AsyncAPI 3.0 specification."""

__all__ = [
    "Server",
    "ServerVariable",
    "Servers",
]

import re

from collections.abc import Iterator

from pydantic import ConfigDict, Field, RootModel, model_validator

from asyncapi3.models.base import ExternalDocumentation, Reference, Tags
from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.bindings import ServerBindingsObject
from asyncapi3.models.helpers import is_null
from asyncapi3.models.security import SecurityScheme


class ServerVariable(ExtendableBaseModel):
    """
    Server Variable Object.

    An object representing a Server Variable for server URL template substitution.

    This object MAY be extended with Specification Extensions.
    """

    enum: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An enumeration of string values to be used if the substitution options "
            "are from a limited set."
        ),
    )
    default: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The default value to use for substitution, and to send, if an alternate "
            "value is not supplied."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An optional description for the server variable. CommonMark syntax MAY "
            "be used for rich text representation."
        ),
    )
    examples: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        description="An array of examples of the server variable.",
    )


class Server(ExtendableBaseModel):
    """
    Server Object.

    An object representing a message broker, a server or any other kind of computer
    program capable of sending and/or receiving data. This object is used to capture
    details such as URIs, protocols and security configuration. Variable substitution
    can be used so that some details, for example usernames and passwords, can be
    injected by code generation tools.
    """

    host: str = Field(
        description=(
            "REQUIRED. The server host name. It MAY include the port. This field "
            "supports Server Variables. Variable substitutions will be made when a "
            "variable is named in {braces}."
        ),
    )
    protocol: str = Field(
        description="REQUIRED. The protocol this server supports for connection.",
    )
    protocol_version: str | None = Field(
        default=None,
        exclude_if=is_null,
        alias="protocolVersion",
        description=(
            "The version of the protocol used for connection. For instance: "
            "AMQP 0.9.1, HTTP 2.0, Kafka 1.0.0, etc."
        ),
    )
    pathname: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The path to a resource in the host. This field supports Server Variables. "
            "Variable substitutions will be made when a variable is named in {braces}."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An optional string describing the server. CommonMark syntax MAY be used "
            "for rich text representation."
        ),
    )
    title: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A human-friendly title for the server.",
    )
    summary: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A short summary of the server.",
    )
    variables: dict[str, ServerVariable | Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map between a variable name and its value. The value is used for "
            "substitution in the server's host and pathname template."
        ),
    )
    security: list[SecurityScheme | Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A declaration of which security schemes can be used with this server. "
            "The list of values includes alternative security scheme objects that "
            "can be used. Only one of the security scheme objects need to be "
            "satisfied to authorize a connection or operation."
        ),
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A list of tags for logical grouping and categorization of servers."
        ),
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation for this server.",
    )
    bindings: ServerBindingsObject | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map where the keys describe the name of the protocol and the values "
            "describe protocol-specific definitions for the server."
        ),
    )


class Servers(RootModel[dict[str, Server | Reference]]):
    r"""
    Servers Object.

    This model validates that all keys match the pattern r"^[A-Za-z0-9_\-]+$",
    values match Reference or Server objects.
    """

    model_config = ConfigDict(
        revalidate_instances="always",
        validate_assignment=True,
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )

    def __iter__(self) -> Iterator[str]:  # type: ignore[override]
        return iter(self.root)

    def __getitem__(self, item: str) -> Server | Reference:
        return self.root[item]

    def __getattr__(self, item: str) -> Server | Reference:
        return self.root[item]

    @model_validator(mode="after")
    def validate_extensions(self) -> "Servers":
        r"""
        Validate that all keys in the input data match the AsyncAPI patterned
        object key pattern.

        Keys must match the regex pattern ^[A-Za-z0-9_\-]+$
        """
        if not self.root:
            return self

        extension_pattern = re.compile(r"^[A-Za-z0-9_\-]+$")
        for field_name in self.root:
            if not extension_pattern.match(field_name):
                raise ValueError(
                    f"Field '{field_name}' does not match patterned object key pattern."
                    " Keys must contain letters, digits, hyphens, and underscores."
                )
        return self



"""AsyncAPI 3.0 Pydantic models."""

__all__ = [
    "AsyncAPI3",
    "Channel",
    "ChannelBindings",
    "ChannelBindingsObject",
    "Channels",
    "Components",
    "Contact",
    "CorrelationID",
    "CorrelationIDs",
    "ExternalDocs",
    "ExternalDocumentation",
    "Info",
    "License",
    "Message",
    "MessageBindings",
    "MessageBindingsObject",
    "MessageExample",
    "MessageTrait",
    "Messages",
    "MultiFormatSchema",
    "OAuthFlow",
    "OAuthFlows",
    "Operation",
    "OperationBindings",
    "OperationBindingsObject",
    "OperationReply",
    "OperationReplyAddress",
    "OperationTrait",
    "OperationTraits",
    "Operations",
    "Parameter",
    "Parameters",
    "PatternedRootModel",
    "Reference",
    "Replies",
    "ReplyAddresses",
    "Schema",
    "Schemas",
    "SecurityScheme",
    "SecuritySchemes",
    "Server",
    "ServerBindings",
    "ServerBindingsObject",
    "ServerVariable",
    "ServerVariables",
    "Servers",
    "Tag",
    "Tags",
]

from asyncapi3.models.asyncapi import AsyncAPI3
from asyncapi3.models.base import ExternalDocumentation, Reference, Tag, Tags
from asyncapi3.models.base_models import PatternedRootModel
from asyncapi3.models.bindings import (
    ChannelBindingsObject,
    MessageBindingsObject,
    OperationBindingsObject,
    ServerBindingsObject,
)
from asyncapi3.models.channel import Channel, Channels, Parameter, Parameters
from asyncapi3.models.components import (
    ChannelBindings,
    Components,
    CorrelationIDs,
    ExternalDocs,
    MessageBindings,
    OperationBindings,
    OperationTraits,
    Replies,
    ReplyAddresses,
    Schemas,
    SecuritySchemes,
    ServerBindings,
    ServerVariables,
)
from asyncapi3.models.info import Contact, Info, License
from asyncapi3.models.message import Message, MessageExample, Messages, MessageTrait
from asyncapi3.models.operation import (
    Operation,
    OperationReply,
    OperationReplyAddress,
    Operations,
    OperationTrait,
)
from asyncapi3.models.schema import MultiFormatSchema, Schema
from asyncapi3.models.security import (
    CorrelationID,
    OAuthFlow,
    OAuthFlows,
    SecurityScheme,
)
from asyncapi3.models.server import Server, Servers, ServerVariable



"""Channel models for AsyncAPI 3.0 specification."""

__all__ = [
    "Channel",
    "Channels",
    "Parameter",
    "Parameters",
]

import re

from collections.abc import Iterator

from pydantic import ConfigDict, Field, RootModel, model_validator

from asyncapi3.models.base import ExternalDocumentation, Reference, Tags
from asyncapi3.models.base_models import ExtendableBaseModel
from asyncapi3.models.bindings import ChannelBindingsObject
from asyncapi3.models.helpers import is_null
from asyncapi3.models.message import Messages


class Parameter(ExtendableBaseModel):
    """
    Parameter Object.

    Describes a parameter included in a channel address.

    This object MAY be extended with Specification Extensions.
    """

    enum: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An enumeration of string values to be used if the substitution options "
            "are from a limited set."
        ),
    )
    default: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "The default value to use for substitution, and to send, if an alternate "
            "value is not supplied."
        ),
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An optional description for the parameter. CommonMark syntax MAY be used "
            "for rich text representation."
        ),
    )
    examples: list[str] | None = Field(
        default=None,
        exclude_if=is_null,
        description="An array of examples of the parameter value.",
    )
    location: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A runtime expression that specifies the location of the parameter value."
        ),
    )


class Parameters(RootModel[dict[str, Parameter | Reference]]):
    r"""
    Parameters Object.

    This model validates that all keys match the pattern r"^[A-Za-z0-9_\-]+$",
    values match Reference or Parameter objects.
    """

    model_config = ConfigDict(
        revalidate_instances="always",
        validate_assignment=True,
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=True,
    )

    def __iter__(self) -> Iterator[str]:  # type: ignore[override]
        return iter(self.root)

    def __getitem__(self, item: str) -> Parameter | Reference:
        return self.root[item]

    def __getattr__(self, item: str) -> Parameter | Reference:
        return self.root[item]

    @model_validator(mode="after")
    def validate_parameter_keys(self) -> "Parameters":
        r"""
        Validate that all keys in the input data match the AsyncAPI patterned
        object key pattern.

        Keys must match the regex pattern ^[A-Za-z0-9_\-]+$
        """
        if not self.root:
            return self

        extension_pattern = re.compile(r"^[A-Za-z0-9_\-]+$")
        for field_name in self.root:
            if not extension_pattern.match(field_name):
                raise ValueError(
                    f"Field '{field_name}' does not match patterned object key pattern."
                    " Keys must contain letters, digits, hyphens, and underscores."
                )
        return self


class Channel(ExtendableBaseModel):
    """
    Channel Object.

    Describes a shared communication channel.

    This object MAY be extended with Specification Extensions.
    """

    # TODO: What is 'unknown' mean?
    address: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An optional string representation of this channel's address. The address "
            "is typically the 'topic name', 'routing key', 'event type', or 'path'. "
            "When null or absent, it MUST be interpreted as unknown. This is useful "
            "when the address is generated dynamically at runtime or can't be known "
            "upfront. It MAY contain Channel Address Expressions. Query parameters "
            "and fragments SHALL NOT be used, instead use bindings to define them."
        ),
    )
    messages: Messages | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map of the messages that will be sent to this channel by any "
            "application at any time. Every message sent to this channel MUST be "
            "valid against one, and only one, of the message objects defined in this "
            "map."
        ),
    )
    title: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A human-friendly title for the channel.",
    )
    summary: str | None = Field(
        default=None,
        exclude_if=is_null,
        description="A short summary of the channel.",
    )
    description: str | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An optional description of this channel. CommonMark syntax can be used "
            "for rich text representation."
        ),
    )
    # TODO: How to deal with deref recommendation?
    servers: list[Reference] | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "An array of $ref pointers to the definition of the servers in which "
            "this channel is available. If the channel is located in the root "
            "Channels Object, it MUST point to a subset of server definitions located "
            "in the root Servers Object, and MUST NOT point to a subset of server "
            "definitions located in the Components Object or anywhere else. "
            "If the channel is located in the Components Object, it MAY point to a "
            "Server Objects in any location. If servers is absent or empty, this "
            "channel MUST be available on all the servers defined in the Servers "
            "Object. Please note the servers property value MUST be an array of "
            "Reference Objects and, therefore, MUST NOT contain an array of "
            "Server Objects. However, it is RECOMMENDED that parsers "
            "(or other software) dereference this property for a better development "
            "experience."
        ),
    )
    parameters: Parameters | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map of the parameters included in the channel address. It MUST be "
            "present only when the address contains Channel Address Expressions."
        ),
    )
    tags: Tags | None = Field(
        default=None,
        exclude_if=is_null,
        description="A list of tags for logical grouping of channels.",
    )
    external_docs: ExternalDocumentation | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        alias="externalDocs",
        description="Additional external documentation for this channel.",
    )
    bindings: ChannelBindingsObject | Reference | None = Field(
        default=None,
        exclude_if=is_null,
        description=(
            "A map where the keys describe the name of the protocol and the values "
            "describe protocol-specific definitions for the channel."
        ),
    )

    @model_validator(mode="after")
    def validate_parameters(self) -> "Channel":
        """
        Validate that parameters are present only when address contains
        Channel Address Expressions.

        Channel Address Expressions are expressions enclosed in curly braces
        like {userId}.
        """
        if self.parameters is None:
            return self

        if self.address is None:
            raise ValueError(
                "parameters must not be provided when address is null or absent"
            )

        # Check if address contains channel address expressions (curly braces)
        expression_pattern = re.compile(r"\{[^}]+\}")
        if not expression_pattern.search(self.address):
            raise ValueError(
                "parameters must not be provided when address does not contain "
                "Channel Address Expressions"
            )

        return self


# Channels is a type alias for a dictionary of Channel objects
Channels = dict[str, Channel | Reference]
